@@ -1,3 +1,5 @@
+static char rcsid[] = "$Id: approve.c,v 1.103 1998/04/07 07:18:31 will Exp $";
+
 /* The code should compile with either ANSI C or K&R compilers. */
 
 /*
@@ -29,39 +31,31 @@
  * New calls to approve() overwrite the buffer containing the returned path.
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-extern char *error_prog;
 char *
-approve(usrcmd, givehelp, giveversion, verbosity)
+approve(usrcmd, givehelp, verbosity)
 char *usrcmd;		/* command we're to check on. Should NOT be a
 			 * command-line option!  Can be a NULL pointer
 			 * or empty string, in which case we give help,
 			 * print version info, or whatever.
 			 */
 int givehelp;		/* Don't execute command; give help instead. */
-int giveversion;	/* Don't execute command; give version info instead. */
 int verbosity;		/* HELP_BASIC, HELP_FULL, or HELP_FACTS */
 {
     char *buf, *bbuf;
     char *do_variables();
     int indentok;
 
-    currfile = file_open(currfile, superfile);
+    /* Do we (1) give help, or (2) match a command with this user? */
+    if (givehelp && !check_syntax)
+	printhelp_hello(verbosity, usrcmd);
+    if (givehelp && verbosity == HELP_USAGE)
+	return "";
+
+    currfile = file_open(currfile, superfile, 0);
     if (!currfile)
 	return NULL;
 
-    /* Do we (1) print version info, (2) give help, or
-     * (3) match a command with this user?
-     */
-
-    if ((givehelp && verbosity != HELP_FACTS) || giveversion)
-	(void) printf("%s version %s patchlevel %s\n",
-				error_prog, Version, Patchlevel);
-    if (giveversion && !givehelp)
-	return "";
-    if (givehelp)
-	printhelp_hello(verbosity, usrcmd);
-
-    for (li.progs.commandfound=NULL, currfile->line=1;
+    for (localinfo.progs.commandfound=NULL, currfile->line=1;
 	(buf=fillbuffer(currfile->fp, &indentok, &currfile->nl)) ||
 	    ( (check_syntax || indentok) && currfile->prev != NULL);
 					     currfile->line += currfile->nl) {
@@ -102,8 +96,11 @@ int verbosity;		/* HELP_BASIC, HELP_FULL, or HELP_FACTS */
 	buf = clean_buf(buf, NULL);
 	if (!buf)
 	    return NULL;
-	if (debug)
+	if (debug) {
+	    fprintf(stderr,
+	    "==============================================================\n");
 	    fprintf(stderr, "Input text: %s", buf);
+	}
     
 	/* Do variable expansion */
 	buf = do_variables(buf);
@@ -148,22 +145,22 @@ int verbosity;		/* HELP_BASIC, HELP_FULL, or HELP_FACTS */
 
 	    default:
 		/* Have an acceptable line, unless die=xxx was used */
-		return li.progs.cmd_file[li.progs.match].File;
+		return localinfo.progs.cmd_file[localinfo.progs.match].File;
 	    }
 	}
     }
     file_close(currfile);
 
-    if (!check_syntax && indentok) {
+    if (!check_syntax && !indentok) {
 	return NULL;
     } else if (givehelp || check_syntax) {
 	return "";
-    } else if (!li.progs.evermatched) {
+    } else if (!localinfo.progs.evermatched) {
 	Error(0, 0, "No such super command as `%-.500s'.\n", usrcmd);
 	return NULL;
     } else {
 	Error(0, 0, "%-.500s - Permission denied to user %s\n",
-							usrcmd, ui.caller);
+					    usrcmd, userinfo.caller.pw_name);
 	return NULL;
     }
     /* NOTREACHED */
@@ -180,7 +177,7 @@ int verbosity;		/* HELP_BASIC, HELP_FULL, or HELP_FACTS */
  *		of ":"-command, etc.);
  *	return 1 if success.
  *	The cmd/file list and which one matched (if any) are returned
- *	through the li.progs struct.
+ *	through the localinfo.progs struct.
  */
 int
 parseline(givehelp, checksyntax, buf, usrcmd)
@@ -196,7 +193,7 @@ char *usrcmd;		/* Command user wants to execute -- can be
     char *word1;
     char *cond_or_opt_wd;
     char *bbuf;
-    int i;
+    int i, continue_after_if;
     int condition_match;
     /* FullPath is twice MAXPATHLEN, because this allows room for
      * MAXPATHLEN characters of arguments as well as MAXPATHLEN
@@ -204,13 +201,13 @@ char *usrcmd;		/* Command user wants to execute -- can be
      */
     static char FullPath[2*MAXPATHLEN];
 
-    li.progs.commandfound = NULL;
+    localinfo.progs.commandfound = NULL;
 
     bbuf = buf;
     buf += strspn(buf, SEP);	/* skip leading whitespace */
     if (!*buf || *buf == '#')
 	return 0;		/* Discard empty lines and pure-comment lines */
-
+    
     if (bbuf != buf) {
 	return Error(0, 0, "%t\n\tformat error in super(1) file: \n\
 \tCommandPatterns must begin in column 1.\n");
@@ -220,14 +217,14 @@ char *usrcmd;		/* Command user wants to execute -- can be
     strqS_qm = my_qm;
     strqS_cc = my_cc;
 
-    if (li.progs.nalloc == 0) {
+    if (localinfo.progs.nalloc == 0) {
 	int nalloc = 10;
-	if ( !(li.progs.cmd_file = (ProgList *)
+	if ( !(localinfo.progs.cmd_file = (ProgList *)
 				malloc(nalloc * sizeof(ProgList))) )
 	    return Error(1, 0, "%tFailed to malloc space for Cmd+File list\n");
-	li.progs.nalloc = 10;
+	localinfo.progs.nalloc = 10;
     }
-    li.progs.n = 0;	/* Initialize to 0 cmd/file pairs */
+    localinfo.progs.n = 0;	/* Initialize to 0 cmd/file pairs */
 
     /* Start by inspecting the first word.  If it begins ":", it is
      * a special builtin command.  Otherwise, if it contains "::", it
@@ -238,15 +235,27 @@ char *usrcmd;		/* Command user wants to execute -- can be
 	return Error(0, 0, "%t\n\tformat error in super.tab file: \
 CmdPattern or FullPathName missing.\n");
 
-    if (*word1 == ':' || strcmp(word1, "/") == 0) {
-	/* It's a special builtin command beginning ":" (or the obsolete
-	 * special command name "/".)
-	 */
-	int i;
+    do {
+	continue_after_if = 0;
+	if (*word1 == ':' || strcmp(word1, "/") == 0) {
+	    /* It's a special builtin command beginning ":" (or the obsolete
+	     * special command name "/".)
+	     */
+	    int i;
 
-	i = process_colon_cmds(word1);
-	return (i < 0) ? -1 : 0;
-    }
+	    i = process_colon_cmds(word1, &continue_after_if);
+	    if (!continue_after_if) {
+		return (i < 0) ? -1 : 0;
+	    } else {
+		/* A :if  evaluated to true; replace "word1" with the next
+		 * word from this line, and process the rest of the line.
+		 */
+		word1 = strqtokS(NULL, SEP, NULL, NULL, 1);
+		if (!word1)
+		    return 0;	/* no body for conditional */
+	    }
+	}
+    } while (continue_after_if);
 
     /* Build lists of cmd::file entries */
     if ((i=build_cmd_file(usrcmd, givehelp || checksyntax,
@@ -255,10 +264,10 @@ CmdPattern or FullPathName missing.\n");
 
 
     if (!(givehelp || checksyntax)) {
-	int indx = (li.progs.match >= 0) ? li.progs.match : 0;
+	int indx = (localinfo.progs.match >= 0) ? localinfo.progs.match : 0;
 
 	/* Check if we are supposed to generate a modified FullPath */
-	char *path = li.progs.cmd_file[indx].File;
+	char *path = localinfo.progs.cmd_file[indx].File;
 
 	/* Adjust fullpath as necessary */
 	if (fixup_fullpath(indx, usrcmd, path,
@@ -267,27 +276,36 @@ CmdPattern or FullPathName missing.\n");
 
     } else {
 	/* givehelp or check syntax */
-	int indx = (li.progs.match >= 0) ? li.progs.match : 0;
-	li.progs.commandfound = li.progs.cmd_file[indx].Cmd;
+	int indx = (localinfo.progs.match >= 0) ? localinfo.progs.match : 0;
+	localinfo.progs.commandfound = localinfo.progs.cmd_file[indx].Cmd;
     }
 
-    if (debug)
-	fprintf(stderr, "\t*** commandfound = `%s' path = `%s' ***\n",
-	    li.progs.commandfound, li.progs.match >= 0 ?
-			li.progs.cmd_file[li.progs.match].File : "-");
+    if (!debug && it_came_from_cmdline) {
+	/* Not debug, but -F / -T / -U / -G / -M flag was given -- so, let's
+	 * print a separator line to help out the reader.
+	 */
+	fprintf(stderr,
+	"==============================================================\n");
+    }
+    if (!givehelp && (debug || it_came_from_cmdline)) {
+	fprintf(stderr, "\tCommandfound = `%s' path = `%s'\n",
+	    localinfo.progs.commandfound, localinfo.progs.match >= 0 ?
+		localinfo.progs.cmd_file[localinfo.progs.match].File : "-");
+    }
 
     /* Now process command-line conditions and options */
     if ((condition_match=conditions_and_options(cond_or_opt_wd)) == -1)
 	return condition_match;
 
 
-    if (condition_match && li.die) {
+    if (condition_match && localinfo.die) {
 	if (checksyntax)
 	    return 0;
 	else if (givehelp)
 	    return -1;
 	else
-	    return Error(0, 0, "User's command: %s. %s\n", usrcmd, li.die);
+	    return Error(0, 0, "User's command: %s. %s\n",
+				    usrcmd, localinfo.die);
     }
     return condition_match;
 }
@@ -324,7 +342,7 @@ char **cond_or_opt_wd;
 	 * the matching filename.
 	 */
 	pair = word1;
-	li.progs.match = -1;	/* initialize to no match */
+	localinfo.progs.match = -1;	/* initialize to no match */
 	do {
 	    s = strchr(pair, ':');
 	    coloncolon = (s && *(s+1) == ':') ? s : NULL;
@@ -334,47 +352,52 @@ char **cond_or_opt_wd;
 		break;
 	    }
 	    if (debug)
-		fprintf(stderr, "Command::file = `%s'\n", pair);
+		fprintf(stderr, "\tCommand::file = `%s'\n", pair);
 
 	    if (coloncolon == pair)
 		return Error(0, 0, "%t\n\tFormat error in super.tab file:\n\
 \tCmd::File used, but the Cmd part is missing.\n");
 	    *coloncolon = '\0';
-	    li.progs.n++;	/* Have a new cmd/file pair */
-	    if (li.progs.n == li.progs.nalloc) {
-		li.progs.nalloc *= 2;
-		li.progs.cmd_file = (ProgList *) realloc(li.progs.cmd_file,
-				li.progs.nalloc * sizeof(ProgList));
-		if (!li.progs.cmd_file)
+	    localinfo.progs.n++;	/* Have a new cmd/file pair */
+	    if (localinfo.progs.n == localinfo.progs.nalloc) {
+		localinfo.progs.nalloc *= 2;
+		localinfo.progs.cmd_file =
+			    (ProgList *) realloc(localinfo.progs.cmd_file,
+				localinfo.progs.nalloc * sizeof(ProgList));
+		if (!localinfo.progs.cmd_file)
 		    return Error(1, 0,
 			"%tFailed to malloc space for Cmd+File list\n");
 	    }
 
-	    li.progs.cmd_file[li.progs.n - 1].Cmd = pair;
-	    li.progs.cmd_file[li.progs.n - 1].File = coloncolon + 2;
+	    localinfo.progs.cmd_file[localinfo.progs.n - 1].Cmd = pair;
+	    localinfo.progs.cmd_file[localinfo.progs.n - 1].File =
+							coloncolon + 2;
 
 	    if ((givehelp_or_checksyntax && !usrcmd) ||
-					match_pattern(0, usrcmd, pair) == 1) {
+				    match_pattern(0, 2, usrcmd, pair) == 1) {
 		/* Matched the command */
-		li.progs.match = li.progs.n - 1;
-		li.progs.evermatched = 1;
-		if (li.progs.cmd_file[li.progs.match].File[0] == '\0')
+		localinfo.progs.match = localinfo.progs.n - 1;
+		localinfo.progs.evermatched = 1;
+		if (localinfo.progs.cmd_file[localinfo.progs.match].File[0]
+				== '\0')
 		    return Error(0, 0, "%t\n\tFormat error in super.tab file:\n\
 \tCmd::File used, but the File part is missing.\n");
 
-		if (gi.relative_path == 0 &&
-			    li.progs.cmd_file[li.progs.match].File[0] != '/')
+		if (globalinfo.relative_path == 0 &&
+		    localinfo.progs.cmd_file[localinfo.progs.match].File[0]
+				!= '/')
 		    return Error(0, 0, "%t\n\tFormat error in super.tab file:\n\
 \tfilename `%s' is not an absolute path.\n\
 \tUse global option relative_path=y if you really want such foolishness!\n",
-		    li.progs.cmd_file[li.progs.match].File);
+		    localinfo.progs.cmd_file[localinfo.progs.match].File);
 	    }
 
 	} while ( (pair = strqtokS(NULL, SEP, NULL, NULL, 1)) );
 	if (!pair)
 	    *cond_or_opt_wd = NULL;
 
-	if (!(givehelp_or_checksyntax && !usrcmd) && li.progs.match == -1)
+	if (!(givehelp_or_checksyntax && !usrcmd) &&
+				localinfo.progs.match == -1)
 	    return 0;		/* no match */
 
     } else if (s && *s == ':') {
@@ -385,32 +408,34 @@ The CmdPattern (%s) may not contain an embedded colon\n", word1);
 
     } else {
 	/* It was a plain command name.  Get the full path. */
-	li.progs.n = 1;
-	li.progs.match = -1;	/* initialize to no match */
-	li.progs.cmd_file[0].Cmd = word1;
-	li.progs.cmd_file[0].File = strqtokS(NULL, SEP, NULL, NULL, 1);
-	if (li.progs.cmd_file[0].File == NULL ||
-				li.progs.cmd_file[0].File[0] == '\0')
+	localinfo.progs.n = 1;
+	localinfo.progs.match = -1;	/* initialize to no match */
+	localinfo.progs.cmd_file[0].Cmd = word1;
+	localinfo.progs.cmd_file[0].File = strqtokS(NULL, SEP, NULL, NULL, 1);
+	if (localinfo.progs.cmd_file[0].File == NULL ||
+				localinfo.progs.cmd_file[0].File[0] == '\0')
 	    return Error(0, 0, "%t\n\tformat error in super.tab file: \
 CmdPattern or FullPathName missing.\n");
 
-	    if (gi.relative_path == 0 && li.progs.cmd_file[0].File[0] != '/')
+	    if (globalinfo.relative_path == 0 &&
+			localinfo.progs.cmd_file[0].File[0] != '/')
 		return Error(0, 0, "%t\n\tFormat error in super.tab file: \
 filename `%s' is not an absolute path.\n\
 \tUse global option relative_path=y if you really want such foolishness!\n",
-		li.progs.cmd_file[0].File);
+		localinfo.progs.cmd_file[0].File);
 
 	if (debug)
-	    fprintf(stderr, "CmdPattern = `%s' %s\n",
-		li.progs.cmd_file[0].Cmd, taglines(1));
+	    fprintf(stderr, "\tCmdPattern = `%s' %s\n",
+		localinfo.progs.cmd_file[0].Cmd, taglines(1));
 
 	if ((givehelp_or_checksyntax && !usrcmd)  ||
-		    match_pattern(0, usrcmd, li.progs.cmd_file[0].Cmd) == 1) {
-	    li.progs.match = 0;	/* success */
-	    li.progs.evermatched = 1;
+	    match_pattern(0, 2, usrcmd, localinfo.progs.cmd_file[0].Cmd) == 1) {
+	    localinfo.progs.match = 0;	/* success */
+	    localinfo.progs.evermatched = 1;
 	}
 
-	if (!(givehelp_or_checksyntax && !usrcmd) && li.progs.match == -1)
+	if (!(givehelp_or_checksyntax && !usrcmd) &&
+					localinfo.progs.match == -1)
 	    return 0;		/* Skip non-matching commands */
 
 	/* The next word has to be ready in advance of the next loop: */
@@ -464,11 +489,11 @@ but the superfile entry\nreads `%s'\n", path);
 	    strncpy(FullPath, path, n);
 	strcpy(FullPath + n, usrcmd);
 	strcat(FullPath, p+1);
-	li.progs.commandfound = usrcmd;
-	li.progs.cmd_file[indx].File = FullPath;
+	localinfo.progs.commandfound = usrcmd;
+	localinfo.progs.cmd_file[indx].File = FullPath;
     } else {
 	/* Regular command */
-	li.progs.commandfound = usrcmd;
+	localinfo.progs.commandfound = usrcmd;
     }
     return 0;
 }
@@ -520,9 +545,18 @@ char *cond_or_opt_wd;	/* First word to process; rest we get with strqtokS */
 		/* Default is to treat it as user~ pattern.
 		 * Insert in user list.
 		 */
+		int i;
+		char *s, **glob;
+
 		invert = *cond_or_opt_wd == '!';
-		if (InsertUserList(invert ? cond_or_opt_wd+1 : cond_or_opt_wd,
-				    &li.userpats, &li.origtext, invert) == -1)
+		s = invert ? cond_or_opt_wd+1 : cond_or_opt_wd;
+
+		/* Do brace globbing */
+		if ((i=globbraces(s, 1, &glob)) != 0)
+		    return Error(0, 0, "%tMissing `%c'.\n", i);
+
+		if (InsertUserList(s, glob,
+		    &localinfo.userpats, &localinfo.origtext, invert) == -1)
 		    return -1;
 	    }
 	} while ( (cond_or_opt_wd = strqtokS(NULL, SEP, NULL, NULL, 1)) );
@@ -533,14 +567,16 @@ char *cond_or_opt_wd;	/* First word to process; rest we get with strqtokS */
     /* Look at condition lists in reverse order, and stop as soon
      * as there is a result.
      */
-    if (gi.userafter.next || li.userpats.next || gi.userbefore.next) {
+    if (globalinfo.userafter.next ||
+	    localinfo.userpats.next ||
+	    globalinfo.userbefore.next) {
 	/* We have user fields to match */
 	matches.user = 0;	/* default is to disallow user */
-	match_ugh_user(&gi.userafter, 1);
+	match_ugh_user(&globalinfo.userafter, 1);
 	if (!matches.user) {
-	    match_ugh_user(&li.userpats, 0);
+	    match_ugh_user(&localinfo.userpats, 0);
 	    if (!matches.user) {
-		match_ugh_user(&gi.userbefore, 1);
+		match_ugh_user(&globalinfo.userbefore, 1);
 	    }
 	}
     } else {
@@ -551,17 +587,19 @@ char *cond_or_opt_wd;	/* First word to process; rest we get with strqtokS */
 	condition_match = 0;
 
     if (condition_match != 0) {
-	if (gi.timeafter.next || li.time.next || gi.timebefore.next) {
+	if (globalinfo.timeafter.next ||
+		localinfo.time.next ||
+		globalinfo.timebefore.next) {
 	    /* We have time fields to match */
 
 	    /* Default depends on whether all time fields are inverted */
 	    matches.time = -1;
 	    matches.allinverted = 1;
-	    matchtime(&ui.ourtime, &gi.timeafter);
+	    matchtime(&userinfo.ourtime, &globalinfo.timeafter);
 	    if (matches.time == -1) {
-		matchtime(&ui.ourtime, &li.time);
+		matchtime(&userinfo.ourtime, &localinfo.time);
 		if (matches.time == -1) {
-		    matchtime(&ui.ourtime, &gi.timebefore);
+		    matchtime(&userinfo.ourtime, &globalinfo.timebefore);
 		}
 	    }
 	    if (matches.time == -1) {
@@ -569,13 +607,13 @@ char *cond_or_opt_wd;	/* First word to process; rest we get with strqtokS */
 		 * deny execution if there were some non-inverting patterns.
 		 */
 		condition_match = matches.allinverted ? 1 : 0;
-		if (debug) {
+		if (debug || it_came_from_cmdline) {
 		    if (matches.allinverted)
-			fputs("\tPermission allowed (= default when \
-all time patterns are !time~xxx)\n", stderr);
+			fputs("\tPermission allowed (default for non-match \
+when all times are !time~xxx)\n", stderr);
 		    else
-			fputs("\tPermission denied (= default when \
->=1 time pattern is non-negated)\n", stderr);
+			fputs("\tPermission denied (default for non-match \
+when >=1 time is non-negated)\n", stderr);
 		}
 	    } else {
 		condition_match = matches.time;
@@ -593,29 +631,69 @@ printhelp_hello(verbosity, usrcmd)
 int verbosity;
 char *usrcmd;
 {
+    static char *usage[] = {
+"    %s command [args...]	\texecutes command [args]",
+"    %s -h			\tprint this usage info, then exits",
+"    %s -V			\tprint version number, then exits",
+"    %s [-H | -f] [MasqueradeOptions]",
+"    				\tlist ok cmds: -H=verbose,-f=just-facts",
+"    %s [-d | -D] [MasqueradeOptions]  cmd [args...]",
+"    				\t...debug mode",
+"    %s -b			\tlists builtin variables, then exits",
+"    %s -c [superfile]	\tjust check syntax of the superfile",
+"",
+"Notes on options:",
+"    -d    ...basic debug mode: show what would be done,",
+"		but don't actually execute a command.",
+"    -D    ...same as -d, plus tell more about variables, plus other details.",
+"    -f -- (just the facts, ma'm) lists the commands you can execute, as:",
+"		    Cmd FullPath [initial args]",
+"	      Useful for scripts that want a list of what you may execute.",
+"    MasqueradeOptions:",
+"	** No command is ever executed when any MasqueradeOption is used.**",
+"	Additionally, some debugging output is enabled when these are used:",
+"	-F superfile -- names the super file to use.",
+"	-G gid -- act as if the invoking user was group gid.",
+"	-M mach -- act as if the command was executed from host mach.",
+"	-T hh[:mm]/day -- act as if this is execution time.",
+"	-U uid -- act as if the invoking user was user uid.",
+};
+
     if (verbosity == HELP_FACTS)
 	return;
 
-    printf("Use:\n\t%s [-d | -D] command [args]\n", prog);
-    printf("    or:\n");
-    printf("\t%s [-h|-H|-f|-V]\n", prog);
-    printf("\t-h=help; -H=long help; -f=just-facts-help; -V=version; -d,-D=debug\n");
-    printf("    or:\n");
-    printf("\t%s -c\n", prog);
-    printf("\t-c=check syntax in super file; don't execute anything.\n\n");
-    if (verbosity) {
+    if (verbosity == HELP_USAGE) {
+	int i;
+	for (i = 0; i < NELEM(usage); i++) {
+	    fprintf(stderr, usage[i], prog);
+	    putc('\n', stderr);
+	}
+	printf("\nSuper.tab file: `%s'\n\n", superfile);
+	return;
+    }
+
+    if (verbosity == HELP_FULL) {
+	printf("(Use %s -h for general usage information.)\n", prog);
+    }
+
+    if (verbosity != HELP_BASIC) {
 	printf("Super.tab file: `%s'\n\n", superfile);
     }
+
     if (usrcmd) {
 	if (verbosity == HELP_BASIC)
-	    (void) printf("Long help on command `%s' for user %s:\n\n",
-				usrcmd, ui.caller);
-	else if (verbosity == HELP_FULL)
 	    (void) printf("Terse help on command `%s' for user %s:\n\n",
-				usrcmd, ui.caller);
+				usrcmd, userinfo.caller.pw_name);
+	else if (verbosity == HELP_FULL)
+	    (void) printf("Long help on command `%s' for user %s:\n\n",
+				usrcmd, userinfo.caller.pw_name);
     } else {
-	(void) printf("Commands available to user %s%s:\n\n", ui.caller,
-		verbosity ? " (use option `-h' for a terse listing)" :
+	if (verbosity) {
+	    (void) printf("===============================================\n");
+	}
+	(void) printf("Commands available to user %s%s:\n\n",
+		userinfo.caller.pw_name,
+		verbosity ? " (use option `-h' for a general usage listing)" :
 		" (use option `-H' for a long-winded listing)");
 	if (!verbosity) {
 	    (void) printf("Command Name    Comments\n");
@@ -636,32 +714,72 @@ int verbosity;
     int i, j, n;
 
     if (verbosity == HELP_BASIC) {
-	if (li.progs.n < 1) {
+	if (localinfo.progs.n < 1) {
 	    ;
-	} else if (li.progs.n == 1) {
-	    printf("%-15s %s\n", li.progs.cmd_file[0].Cmd,
-					    li.info ? li.info : "");
+	} else if (localinfo.progs.n == 1) {
+	    /* The entry contains a single CmdPattern/FullPath pairs */
+
+	    /* If the command contains a wildcard, print the full filename,
+	     * otherwise just the command name.
+	     */
+	    if (strchr(localinfo.progs.cmd_file[0].Cmd, '*')) {
+		printf("%s -> %s:\n                -- %s\n",
+			localinfo.progs.cmd_file[0].Cmd,
+			localinfo.progs.cmd_file[0].File,
+			localinfo.info);
+	    } else {
+		printf("%-15s %s\n", localinfo.progs.cmd_file[0].Cmd,
+			localinfo.info ? localinfo.info : "");
+	    }
 	} else {
-	    printf("%s", li.progs.cmd_file[0].Cmd);
-	    for (i=1; i<li.progs.n; i++)
-		printf(", %s", li.progs.cmd_file[i].Cmd);
-	    if (li.info)
-		(void) printf(":\n                -- %s\n", li.info);
-	    else
-		(void) putchar('\n');
+	    /* The entry contains multiple CmdPattern/FullPath pairs */
+
+	    /* If any command contains a wildcard, print the
+	     * full filenames, otherwise just the command name.
+	     */
+	    int havewildcard=0;
+	    for (i=0; havewildcard==0 && i<localinfo.progs.n; i++) {
+		havewildcard =
+			(strchr(localinfo.progs.cmd_file[0].Cmd, '*') != NULL);
+	    }
+	    if (havewildcard) {
+		for (i=0; i<localinfo.progs.n-1; i++) {
+		    printf("%s -> %s;\n", prog,
+			localinfo.progs.cmd_file[i].Cmd,
+			localinfo.progs.cmd_file[i].File);
+		}
+		i = localinfo.progs.n - 1;
+		printf("%s -> %s:\n", prog,
+		    localinfo.progs.cmd_file[i].Cmd,
+		    localinfo.progs.cmd_file[i].File);
+		if (localinfo.info && *localinfo.info)
+		    (void) printf("                -- %s\n", localinfo.info);
+
+	    } else {
+		/* Now wildcards in names */
+		printf("%s", localinfo.progs.cmd_file[0].Cmd);
+		for (i=1; i<localinfo.progs.n; i++)
+		    printf(", %s", localinfo.progs.cmd_file[i].Cmd);
+		if (localinfo.info && *localinfo.info)
+		    (void) printf(":\n                -- %s\n",
+			localinfo.info);
+		else
+		    (void) putchar('\n');
+	    }
 	}
 	return;
 
     } else if (verbosity == HELP_FACTS) {
-	for (i=0; i<li.progs.n; i++)
-	    (void) printf("%s %s\n", li.progs.cmd_file[i].Cmd,
-						li.progs.cmd_file[i].File);
+	for (i=0; i<localinfo.progs.n; i++)
+	    (void) printf("%s %s\n", localinfo.progs.cmd_file[i].Cmd,
+					localinfo.progs.cmd_file[i].File);
 	return;
 
     } else if (verbosity == HELP_FULL) {
-	for (i=0; i<li.progs.n; i++)
-	    (void) printf("%s %s -> %s\n", prog, li.progs.cmd_file[i].Cmd,
-						li.progs.cmd_file[i].File);
+	for (i=0; i<localinfo.progs.n; i++)
+	    (void) printf("%s %s -> %s\n", prog,
+			localinfo.progs.cmd_file[i].Cmd,
+			localinfo.progs.cmd_file[i].File);
     } else {
 	(void) Error(0, 0, "%s: printhelp(): Unknown verbosity level %d\n",
 			verbosity);
@@ -669,27 +787,29 @@ int verbosity;
     }
 
     /* Assert verbosity == HELP_FULL */
-    if (li.info)
-	printf("\t## %s\n", li.info);
+    if (localinfo.info)
+	printf("\t## %s\n", localinfo.info);
 
-    if (li.die) {
+    if (localinfo.die) {
 	printf("\tCommand forces exit using 'die' option:\n\t\tdie=\"%s\"",
-		li.die);
+		localinfo.die);
 	return;
     }
 
-    if (li.print)
+    if (localinfo.print)
 	printf("\tCommand prints msg before execution:\n\t\t\"%s\"",
-		li.print);
+		localinfo.print);
 
-    if (gi.timebefore.next || gi.timeafter.next || li.time.next) {
+    if (globalinfo.timebefore.next ||
+	    globalinfo.timeafter.next ||
+	    localinfo.time.next) {
 	printf("\tPermitted times for execution (in reverse input order):\n");
 	for (j = 0; j<3; j++) {
 	    TimeList *tl;
 	    switch (j) {
-	    case 0: tl = gi.timebefore.next; break;
-	    case 1: tl = li.time.next; break;
-	    default: tl = gi.timeafter.next; break;
+	    case 0: tl = globalinfo.timebefore.next; break;
+	    case 1: tl = localinfo.time.next; break;
+	    default: tl = globalinfo.timeafter.next; break;
 	    }
 	    for ( ; tl; tl=tl->next) {
 		printf("\t\t%stime~%d:%02d-%d:%02d/%s\n",
@@ -699,34 +819,56 @@ int verbosity;
 	    }
 	}
     }
-    if (*li.user || *li.group || *li.u_g || li.fdlist || *li.owner || li.env) {
+    if (*localinfo.user ||
+	    *localinfo.group ||
+	    *localinfo.u_g ||
+	    localinfo.fdlist ||
+	    *localinfo.owner ||
+	    localinfo.env) {
 	(void) fputs("\t(Executes with:", stdout);
-	if (*li.user) (void) printf(" uid=%s", li.user);
-	if (*li.group) (void) printf(" gid=%s", li.group);
-	if (*li.u_g) (void) printf(" u+g=%s", li.u_g);
-	if (li.env) (void) printf(" env=%s", li.env);
-	for (p=li.setenv; *p; p++)
-	    (void) printf(" setenv=%s", *p);
-	if (li.fdlist) (void) printf(" fdlist=%s", li.fdlist);
-	if (*li.owner) (void) printf(" owner=%s", li.owner);
+	if (*localinfo.user) (void) printf(" uid=%s", localinfo.user);
+	if (*localinfo.group) (void) printf(" gid=%s", localinfo.group);
+	if (*localinfo.u_g) (void) printf(" u+g=%s", localinfo.u_g);
+	if (localinfo.env) {
+	    for (p=localinfo.env; *p; p++)
+		(void) printf("%s%s", p==localinfo.env ? " env=" : ",", *p);
+	}
+	if (localinfo.setenv) {
+	    for (p=localinfo.setenv; *p; p++)
+		(void) printf(" setenv=%s", *p);
+	}
+	if (localinfo.fdlist) (void) printf(" fdlist=%s", localinfo.fdlist);
+	if (*localinfo.owner) (void) printf(" owner=%s", localinfo.owner);
 	(void) fputs(")\n", stdout);
     }
-    if (li.usr_args[0] < 0)
+
+    if (localinfo.usr_args[0] < 0)
 	;
-    else if (li.usr_args[0] == li.usr_args[1] &&
-					    li.usr_args[0] == 0)
+    else if (localinfo.usr_args[0] == localinfo.usr_args[1] &&
+					    localinfo.usr_args[0] == 0)
 	(void) printf("\tNo user-entered args are allowed.\n");
-    else if (li.usr_args[0] == li.usr_args[1])
+    else if (localinfo.usr_args[0] == localinfo.usr_args[1])
 	(void) printf("\t%d user-entered arg%s required.\n",
-	    li.usr_args[0], li.usr_args[0] == 1? " is" : "s are");
+	    localinfo.usr_args[0], localinfo.usr_args[0] == 1? " is" : "s are");
     else
 	(void) printf("\t%d - %d user-entered args are required.\n",
-	    li.usr_args[0], li.usr_args[1]);
-    n = StrLastInUse(&li.argpats);
+	    localinfo.usr_args[0], localinfo.usr_args[1]);
+
+    if (localinfo.maxlen1arg < 0)
+	(void) printf("\tNo per-arg length limit; ");
+    else
+	(void) printf("\tMax per-arg length: %d chars; ", localinfo.maxlen1arg);
+
+    if (localinfo.maxlenargs < 0)
+	(void) printf("no limit on sum of all args.\n");
+    else
+	(void) printf("max over all args: %d chars.\n", localinfo.maxlenargs);
+
+    n = StrLastInUse(&localinfo.argpats);
     if (n >= 0) {
 	printf("\tUser-entered arguments must match following patterns:\n");
 	for (i = 0; i<=n; i++) {
-	    s = StrEltGetPtr(&li.argpats, i);
+	    s = StrEltGetPtr(&localinfo.argpats, i);
 	    if (s)
 		printf("\t\tArg %d must match pattern: %s\n", i, s);
 	}
@@ -735,16 +877,16 @@ int verbosity;
     if (rcl_nice_incr() != 0)
 	printf("\tCommand executes with nice increment = %d.\n",
 		    rcl_nice_incr());
-    if (rcl_umask() != ui.orig_mask)
+    if (rcl_umask() != userinfo.orig_mask)
 	printf("\tCommand executes with umask set to 0%o.\n", rcl_umask());
 
 
-    if (li.passinfo.required) {
-	if (li.passinfo.timeout == 0)
+    if (localinfo.passinfo.required) {
+	if (localinfo.passinfo.timeout == 0)
 	    (void) printf("\tPassword is required on each use.\n");
 	else
 	    (void) printf("\tPassword is required; good for %d minutes%s.\n",
-		li.passinfo.timeout, li.passinfo.renewtime ? 
+		localinfo.passinfo.timeout, localinfo.passinfo.renewtime ? 
 		    ", extended with each use" : "");
     }
     (void) putchar('\n');
