@@ -1,3 +1,4 @@
+static char rcsid[] = "$Id: checks.c,v 1.52 1998/04/07 08:55:16 will Exp $";
 /* The code should compile with either ANSI C or K&R compilers. */
 
 /*
@@ -10,11 +11,8 @@
 
 #include "super.h"
 #include "version.h"
-#ifdef __linux__
-#include <shadow.h>
-#endif
 
-#ifdef USE_NETGROUP
+#ifdef HAVE_INNETGR
 #define netgrp_u_compare(pattern, user) innetgr(pattern, NULL, user, NULL)
 #define netgrp_h_compare(pattern, host) innetgr(pattern, host, NULL, NULL)
 #else
@@ -41,14 +39,14 @@ char *pat;	/* pattern that value must match */
 	    "\tcheckenv args: name=\"%s\"; value=\"%s\"; pat=\"%s\"\n",
 	    name, value, pat);
     
-    /* Environment variables are always checked with re_comp/re_exec:
+    /* Environment variables are always checked with s_re_comp/s_re_exec:
      * the patterns are fixed internally, not supplied by the user.
      */
-    if (re_comp(pat))
+    if (s_re_comp(pat))
 	return Error(0, 0,
 	    "%t\n\tcheckenv(): couldn't compile pattern `%-.500s'.\n", pat);
 
-    if (re_exec(value) != 1)
+    if (s_re_exec(value) != 1)
 	return Error(0, 0,
 	    "checkenv(): $%.100s (=%.100s) doesn't match pattern %-.500s.\n",
 			name, value, pat);
@@ -56,130 +54,242 @@ char *pat;	/* pattern that value must match */
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Check ownership of the file */
+/* Look up owner of a file, return uid and gid of owner */
+/* Return 0 on success, -1 & print message on failure */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-check_owner(file)
+get_owner(file, uid_p, gid_p)
 char *file;
+uid_t *uid_p;
+gid_t *gid_p;
 {
-    /* Return 0 if file ownership ok (or if no check required); -1 if not ok */
+    /* Return 0 if file ownership ok; -1 if not ok */
     struct stat st;
-    struct passwd *owner_pw;
 
     if (!file || *file == '\0')
-	return Error(1, 0, "check_owner(): passed null ptr or empty string\n");
-
-    if (*li.owner == '\0')
-	return 0;	/* no checks required */
+	return Error(0, 0, "get_owner(): passed null ptr or empty string\n");
 
     if (stat(file, &st) == -1)
 	return Error(1, 0, "stat() failed on file `%s': ", file);
 
-    /* Convert owner string to a uid */
-    owner_pw = getpwnam(li.owner);
-    if (!owner_pw) {
-	/* Maybe owner was a uid, not a username */
-	char c;
-	int numeric, i;
-	numeric = (sscanf(li.owner, "%d%c", &i, &c) == 1);
-	if (numeric)
-	    owner_pw = getpwuid(i);
-    }
-    if (!owner_pw)
-	return Error(0, 0,
-	"Can't check owner=%s: no such user or uid as `%s' in password file.\n",
-	li.owner);
+    *uid_p = st.st_uid;
+    *gid_p = st.st_gid;
 
-    if (st.st_uid != owner_pw->pw_uid)
+    return 0;
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Check ownership of the file */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+int
+check_owner()
+{
+    /* Return 0 if file ownership ok; -1 if not ok */
+    struct passwd *owner_pw;
+
+    if (*localinfo.owner == '\0')
+	return 0;	/* no checks required */
+
+    /* Convert desired-owner string to a uid */
+    owner_pw = getpwentry(1, localinfo.owner);
+    if (!owner_pw)
+	return -1;
+
+    if (localinfo.file_uid != owner_pw->pw_uid)
 	return Error(0, 0,
 	    "Actual owner of `%s' is uid %d, but superfile \
-requires owner to be %d (%s).\n", file, st.st_uid,
-	    owner_pw->pw_uid, owner_pw->pw_name);
+requires owner to be %d (%s).\n",
+	    localinfo.progs.cmd_file[localinfo.progs.match].File,
+	    localinfo.file_uid, owner_pw->pw_uid, owner_pw->pw_name);
     
     return 0;
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Set user and group according to the specified args */
+/* Gets an entry from the password file.  Optionally accepts special bracketed
+ * name (e.g. <owner> or <caller>).  Accepts text uid's instead of names.
+ * Returns ptr to password entry.  The password entry may be that returned
+ * by getpwnam(), or it may be one already stored in a super-owned struct;
+ * you can't make any assumptions about it -- therefore don't modify it.
+ *
+ * The returned pointer points to an area that may be overwritten by later
+ * calls to the getpwxxx() routines; therefore if the caller wants to save
+ * the data, the data must be copied.
+ *
+ * On error, print message and return NULL.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+struct passwd *
+getpwentry(allow_brackets, username)
+int allow_brackets;	/* allow special names like <caller> */
+char *username;		/* name to translate */
+{
+    struct passwd *pw;
+    int l = strlen(username);
+
+    if (allow_brackets && username[0] == '<' && username[l-1] == '>') {
+	/* Translate special name */
+	if (strcmp(username, "<owner>") == 0) {
+	    if (localinfo.file_uid != UID_NOTSET) {
+		pw = getpwuid(localinfo.file_uid);
+	    } else {
+		Error(0, 0,
+	    "%t: getpwentry() Internal Error!\n\tFile owner not yet known!\n");
+		return NULL;
+	    }
+	} else if (strcmp(username, "<caller>") == 0) {
+	    pw = &userinfo.caller;
+	} else {
+	    Error(0, 0, "%t\n\t\tUnknown special name %s\n", username);
+	    return NULL;
+	}
+    } else {
+	/* Regular name or number */
+	pw = getpwnam(username);
+	if (!pw) {
+	    char c;
+	    int i, numeric;
+	    numeric = (sscanf(username, "%d%c", &i, &c) == 1);
+	    if (numeric)
+		pw = getpwuid(i);
+	}
+    }
+    if (!pw) {
+	Error(0, 0, "%t\n\tNo such user or uid as `%s' in password file.\n",
+	    username);
+	return NULL;
+    }
+    return pw;
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Set supplementary groups according to the specified args                */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+#ifndef HAVE_GETGROUPS
+int
+set_suppl_groups() { return 0; }
+#else
+int
+set_suppl_groups()
+{
+    GETGROUPS_T *addgroups_p, *groups_p, groups[NGROUPS_MAX];
+    GETGROUPS_T gotten_groups[NGROUPS_MAX];
+    int naddgroups, ngroups;
+    int i;
+
+    /* Set the supplementary groups */
+    if (*localinfo.u_g) {
+	/* Default is to take from that user.  (Don't worry if there is no
+	 * such user; it'll get noticed later.)  Conflicts with local
+	 * option groups=xxx, which is disallowed.
+	 */
+	if (localinfo.ngroups != GROUPS_NOTSET && !localinfo.groups_added)
+	    return Error(0, 0, "%t\n\t\tu+g=xxx conflicts with groups=yyy \
+and may not be used in the same entry\n");
+
+	initgroups(localinfo.user, userinfo.new_gid);
+	/* Check for local or global addgroups */
+	if (localinfo.ngroups != GROUPS_NOTSET) {
+	    addgroups_p = localinfo.groups;
+	    naddgroups = localinfo.ngroups;
+	} else if (globalinfo.ngroups != GROUPS_NOTSET &&
+					globalinfo.groups_added) {
+	    addgroups_p = globalinfo.groups;
+	    naddgroups = globalinfo.ngroups;
+	} else {
+	    addgroups_p = NULL;
+	    naddgroups = 0;
+	}
+	ngroups = Getgroups(NGROUPS_MAX, gotten_groups);
+	if (ngroups == -1)
+	    return Error(1, 0, "%t Getgroups() failed: ");
+	if (ngroups + naddgroups > NGROUPS_MAX)
+	    return Error(1, 0,
+			    "%t\n\t\taddgroups=xxx adds too many groups.");
+	for (i = 0; i < ngroups; i++)
+	    groups[i] = gotten_groups[i];
+	for (groups_p = &groups[ngroups], i=0; i < naddgroups; i++)
+	    *groups_p++ = *addgroups_p++;
+	ngroups += naddgroups;
+	    
+    }  else if (localinfo.ngroups != GROUPS_NOTSET) {
+	/* There are some explicit local groups=xxx */
+	if (Setgroups(localinfo.ngroups, localinfo.groups) == -1)
+	    return Error(1, 0, "Failed to set supplementary groups list: ");
+
+    } else if (globalinfo.ngroups != GROUPS_NOTSET) {
+	/* There is an explicit global groups=xxx or addgroups=xxx */
+	if (Setgroups(globalinfo.ngroups, globalinfo.groups) == -1)
+	    return Error(1, 0, "Failed to set supplementary groups list: ");
+    } else {
+	/* Default is no supplementary groups */
+	if (Setgroups(0, localinfo.groups) == -1)
+	    return Error(1, 0,
+			    "Failed to clear supplementary groups list: ");
+    }
+    return 0;
+}
+#endif
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Set user, group, and supplementary groups according to the specified args */
+/* Side effect: if localinfo.user is numeric, change it to a name */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
 set_u_g()
 {
     /* Return 0 on success, -1 on failure */
-    struct passwd *uid_pw, *pw;
-    struct group *gp;
+    struct passwd *uid_pw;
     SETGRENT_T setgrent();
     void endgrent();
-    int i, found_gid, numeric=0;
+    int i, found_gid;
 
-    ui.new_uid = ui.orig_uid;
-    ui.new_gid = ui.orig_gid;
+    userinfo.new_uid = userinfo.caller.pw_uid;
+    userinfo.new_gid = userinfo.caller.pw_gid;
 
     /* Check gid=xxx */
-    if (*li.group) {
-	char c;
-	numeric = (sscanf(li.group, "%d%c", &i, &c) == 1);
-	ui.new_gid = i;
+    if (*localinfo.group) {
+	found_gid = findgid(1, localinfo.group);
 
-	setgrent();
-	for (found_gid=0, gp = getgrent(); gp; gp = getgrent()) {
-	    if ((strcmp(li.group, gp->gr_name) == 0) ||
-		(numeric && ui.new_gid == gp->gr_gid)) {
-		/* Found the gid in the group file */
-		ui.new_gid = gp->gr_gid;
-		found_gid = 1;
-		break;
-	    }
-	}
-	endgrent();
-	if (!found_gid && numeric) {
-	    /* Look through the password file */
-	    for (found_gid=0, pw = getpwent(); pw; pw = getpwent()) {
-		if (ui.new_gid == pw->pw_gid) {
-		    found_gid = 1;
-		    break;
-		}
-	    }
-	    endpwent();
-	}
-	if (!found_gid)
+	if (found_gid == -1)
 	    return Error(0, 0,
-	    "%t\n\tCan't set gid: \
-no such group or gid as `%s' in group%s file.\n",
-		li.group, numeric ? " or password" : "");
+		"%t\n\tCan't set gid: no such group as `%s' in group file.\n",
+		localinfo.group);
+	userinfo.new_gid = found_gid;
     }
 
     /* Check uid=xxx u+g=yyy */
-    if (*li.user || *li.u_g) {
-	if (!*li.user)
-	    strcpy(li.user, li.u_g);
-	uid_pw = getpwnam(li.user);
-	if (!uid_pw) {
-	    char c;
-	    numeric = (sscanf(li.user, "%d%c", &i, &c) == 1);
-	    ui.new_uid = i;
-	    if (numeric)
-		uid_pw = getpwuid(ui.new_uid);
+    if (*localinfo.user || *localinfo.u_g) {
+	if (*localinfo.u_g)
+	    strcpy(localinfo.user, localinfo.u_g);
+	
+	uid_pw = getpwentry(1, localinfo.user);
+	if (!uid_pw)
+	    return -1;
+	userinfo.new_uid = uid_pw->pw_uid;
+	if (*localinfo.u_g) {
+	    userinfo.new_gid = uid_pw->pw_gid;
 	}
-	if (uid_pw) {
-	    ui.new_uid = uid_pw->pw_uid;
-	    if (*li.u_g)
-		ui.new_gid = uid_pw->pw_gid;
-	} else {
-	    return Error(0, 0, "%t\n\tCan't set uid: \
-no such user or uid as `%s' in password file.\n", li.user);
+	if (strcmp(uid_pw->pw_name, localinfo.user) != 0) {
+	    /* localinfo.user must be numeric; convert it to string */
+	    strcpy(localinfo.user, uid_pw->pw_name);
 	}
     }
 
-    /* Now set uid & gid */
-    if (*li.group || *li.u_g)
-	if (setgid(ui.new_gid) == -1)
-	    return Error(1, 0, "setgid(gid=%d) failed: ", ui.new_gid);
-    if (*li.user || *li.u_g)
-	if (setuid(ui.new_uid) == -1)
-	    return Error(1, 0, "setuid(uid=%d) failed: ", ui.new_uid);
+    /* Set supplementary groups */
+    if (set_suppl_groups() == -1)
+	return -1;
 
-    if (li.passinfo.required) {
+
+    /* Now set uid & gid */
+    if (*localinfo.group || *localinfo.u_g)
+	if (setgid(userinfo.new_gid) == -1)
+	    return Error(1, 0, "setgid(gid=%d) failed: ", userinfo.new_gid);
+    if (*localinfo.user || *localinfo.u_g)
+	if (setuid(userinfo.new_uid) == -1)
+	    return Error(1, 0, "setuid(uid=%d) failed: ", userinfo.new_uid);
+
+    if (localinfo.passinfo.required) {
 	/* Get caller's encrypted password */
 	i = get_encrypted_pw();
 	if (i != 0)
@@ -190,7 +300,7 @@ no such user or uid as `%s' in password file.\n", li.user);
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Puts the encrypted password in ui.encr, and the salt in ui.salt.
+/* Puts the encrypted password in userinfo.encr, and the salt in userinfo.salt.
  * Returns 0 on success, -1 on failure to obtain the password.
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -201,13 +311,13 @@ get_encrypted_pw()
 {
     /* Shadow passwords are always used on Sunos 5.x */
     struct spwd *caller_pw;
-    if (!(caller_pw = getspnam(ui.caller)))
+    if (!(caller_pw = getspnam(userinfo.caller.pw_name)))
 	return Error(1, 0,
 	    "Failed to obtain shadow password entry for user %s: ",
-	    ui.caller);
-    strcpy(ui.encr, caller_pw->sp_pwdp);
-    strncpy(ui.salt, caller_pw->sp_pwdp, 2);
-    ui.salt[2] = '\0';
+	    userinfo.caller.pw_name);
+    strcpy(userinfo.encr, caller_pw->sp_pwdp);
+    strncpy(userinfo.salt, caller_pw->sp_pwdp, 2);
+    userinfo.salt[2] = '\0';
     return 0;
 }
 
@@ -226,21 +336,19 @@ get_encrypted_pw()
     if (stat("/.secure/etc/passwd", &st) == 0) {
 	/* Shadow password file exists; use it */
 	struct s_passwd *caller_pw;
-	if (!(caller_pw = getspwnam(ui.caller)))
+	if (!(caller_pw = getspwnam(userinfo.caller.pw_name)))
 	    return Error(1, 0,
 		"Failed to obtain shadow password entry for user %s: ",
-		ui.caller);
-	strcpy(ui.encr, caller_pw->pw_passwd);
-	strncpy(ui.salt, caller_pw->pw_passwd, 2);
-	ui.salt[2] = '\0';
+		userinfo.caller);
+	strcpy(userinfo.encr, caller_pw->pw_passwd);
+	strncpy(userinfo.salt, caller_pw->pw_passwd, 2);
+	userinfo.salt[2] = '\0';
     } else {
 
 	/* Fall through to regular password file lookup. */
-	if (!(caller_pw = getpwnam(ui.caller)))
-	    return Error(0, 0, "No password entry for user %s.\n", ui.caller);
-	strcpy(ui.encr, caller_pw->pw_passwd);
-	strncpy(ui.salt, caller_pw->pw_passwd, 2);
-	ui.salt[2] = '\0';
+	strcpy(userinfo.encr, userinfo.caller.pw_passwd);
+	strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
+	userinfo.salt[2] = '\0';
     }
     return 0;
 }
@@ -251,29 +359,55 @@ int
 get_encrypted_pw()
 {
     struct passwd *caller_pw;
-    struct spwd *caller_sw;
+    struct spwd *caller_ps;
 
-    if (!(caller_pw = getpwnam(ui.caller)))
-	return Error(0, 0, "No password entry for user %s.\n", ui.caller);
+    if (!(caller_pw = getpwnam(userinfo.caller.pw_name)))
+	return Error(0, 0, "No password entry for user %s.\n",
+		userinfo.caller.pw_name);
 
     /* SCO 3.2v4 has "x" in password field to indicate shadow password
      * file has to be consulted.
      */
-    if (strcmp(caller_pw->sp_pwdp, "x") == 0) {
+    if (strcmp(caller_pw->pw_passwd, "x") == 0) {
 	/* Shadow password in use... */
-	if (!(caller_sw = getspnam(ui.caller)))
+	if (!(caller_ps = getspnam(userinfo.caller.pw_name)))
 	    return Error(1, 0,
 	    "Failed to obtain shadow password entry for user %s: ",
-	    ui.caller);
-	strcpy(ui.encr, caller_sw->sp_pwdp);
-	strncpy(ui.salt, caller_sw->sp_pwdp, 2);
-	ui.salt[2] = '\0';
+	    userinfo.caller.pw_name);
+	strcpy(userinfo.encr, caller_ps->sp_pwdp);
+	strncpy(userinfo.salt, caller_ps->sp_pwdp, 2);
+	userinfo.salt[2] = '\0';
 
     } else {
 	/* Fall through to regular password file lookup. */
-	strcpy(ui.encr, caller_pw->pw_passwd);
-	strncpy(ui.salt, caller_pw->pw_passwd, 2);
-	ui.salt[2] = '\0';
+	strcpy(userinfo.encr, caller_pw->pw_passwd);
+	strncpy(userinfo.salt, caller_pw->pw_passwd, 2);
+	userinfo.salt[2] = '\0';
+    }
+    return 0;
+}
+
+#else
+#ifdef __linux__
+
+int
+get_encrypted_pw()
+{
+    /* Use /etc/shadow if it exists; else fall back on std */
+    static struct stat st;
+    if (stat("/etc/shadow", &st) == 0) {
+	struct spwd *spwd = 0L;
+	if (!(spwd = getspnam(userinfo.caller.pw_name)))
+	    return Error(1, 0,
+		"Failed to obtain shadow password entry for user %s: ",
+		userinfo.caller.pw_name);
+	userinfo.caller.pw_passwd = spwd->sp_pwdp;
+	strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
+	userinfo.salt[2] = '\0';
+    } else {
+	strcpy(userinfo.encr, userinfo.caller.pw_passwd);
+	strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
+	userinfo.salt[2] = '\0';
     }
     return 0;
 }
@@ -283,29 +417,19 @@ get_encrypted_pw()
 int
 get_encrypted_pw()
 {
+fprintf(stderr, "here2\n");
     /* Vanilla password file lookup */
-    struct passwd *caller_pw;
-#ifdef __linux__
-    struct  spwd    *spwd = 0L;
-#endif
 
-    if (!(caller_pw = getpwnam(ui.caller)))
-	return Error(0, 0, "No password entry for user %s.\n", ui.caller);
-
-#ifdef __linux__
-    if ((spwd = getspnam(ui.caller)))
-        caller_pw->pw_passwd = spwd->sp_pwdp;
-#endif
-
-    strcpy(ui.encr, caller_pw->pw_passwd);
-    strncpy(ui.salt, caller_pw->pw_passwd, 2);
-    ui.salt[2] = '\0';
+    strcpy(userinfo.encr, userinfo.caller.pw_passwd);
+    strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
+    userinfo.salt[2] = '\0';
     return 0;
 }
 
 #endif
 #endif
 #endif
+#endif
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -331,9 +455,9 @@ char *cmd;
     int l, istat, file_exists, got_pass;
     int timed_out = 0;
     int status;
-    PID_T child;
+    pid_t pid, child;
 
-    if (!li.passinfo.required)
+    if (!localinfo.passinfo.required)
 	return 0;			/* don't need password */
 
     /* Create or update the timestamp file even if the lifetime is 0
@@ -348,8 +472,19 @@ char *cmd;
 
     } else if (child > 0) {
 	/* In parent -- wait to see if the child succeeded */
-	if (wait(&status) < 0)
-	    return Error(1, 0, "wait() on timestamp creation process failed: ");
+	while ((pid = wait(&status)) > 0 && pid != child) {
+	    if (pid == globalinfo.log.pid) {
+		Error(0, 0,
+	    "Warning: logging process died -- logging to file has stopped.");
+		globalinfo.log.pid = -1;
+	    } else {
+		Error(0, 0,
+		    "wait() surprised! check_pass() received pid %d;\n\t\
+expected child pid = %d; waiting for correct pid...\n", pid, child);
+	    }
+	}
+	if (pid == -1)    
+	    return Error(1, 0, "waiting for timestamp creation process: ");
 	else if (status == 0)
 	    return 0;
 	else
@@ -357,19 +492,20 @@ char *cmd;
 	
     } else {
 	/* In child.  setuid, then make and/or test the directory */
-	if (*li.passinfo.user != '\0') {
-	    stringcopy(li.user, li.passinfo.user, sizeof(li.user));
-	    *li.group = '\0';
-	    *li.u_g = '\0';
+	if (*localinfo.passinfo.user != '\0') {
+	    stringcopy(localinfo.user, localinfo.passinfo.user,
+						sizeof(localinfo.user));
+	    *localinfo.group = '\0';
+	    *localinfo.u_g = '\0';
 	    if (set_u_g() == -1) {
 		return Error(1, 0,
 		"failed to setuid %s before setting timestamp file: ",
-		li.user);
+		localinfo.user);
 	    }
 	}
 	/* Make the timestamp directory name */
-	if (!makedirname(gi.passinfo.perhost ? TIMESTAMP_DIR : "",
-							ui.hostname, file))
+	if (!makedirname(globalinfo.passinfo.perhost ? TIMESTAMP_DIR : "",
+						userinfo.hostname, file))
 	    return -1;
 
 	/* Make the timestamp directory */
@@ -377,13 +513,13 @@ char *cmd;
 	    return -1;
 
 	/* Make the file in the timestamp directory */
-	l = strlen(file) + 1 + strlen(ui.caller);
+	l = strlen(file) + 1 + strlen(userinfo.caller.pw_name);
 	if (l >= MAXPATHLEN)
 	    return Error(1, 0,
 	    "Can't create timestamp file: it would exceed MAXPATHLEN = %d\n",
 	    MAXPATHLEN);
 	strcat(file, "/");
-	strcat(file, ui.caller);
+	strcat(file, userinfo.caller.pw_name);
 
 	istat = stat(file, &st);
 	if (istat != 0 && errno != ENOENT)
@@ -391,12 +527,14 @@ char *cmd;
 
 	file_exists = (istat == 0);
 	if (file_exists)
-	    timed_out = (li.passinfo.timeout < 1) ||
-			((time(NULL)-st.st_mtime) > li.passinfo.timeout*60);
+	    timed_out = (localinfo.passinfo.timeout < 1) ||
+			((time(NULL)-st.st_mtime) >
+			localinfo.passinfo.timeout*60);
 
 	got_pass=0;
 	if (!file_exists || timed_out) {
-	    got_pass = (get_password(cmd, ui.caller, ui.salt, ui.encr) == 1);
+	    got_pass = (get_password(cmd, userinfo.caller.pw_name,
+					userinfo.salt, userinfo.encr) == 1);
 	    if (!got_pass)
 		return -1;
 	}
@@ -412,7 +550,7 @@ char *cmd;
 
 	 * Otherwise create the file (unlink it first if it exists).
 	 */
-	if (file_exists && !(got_pass || li.passinfo.renewtime)) {
+	if (file_exists && !(got_pass || localinfo.passinfo.renewtime)) {
 	    if (st.st_uid != geteuid())
 		return Error(0, 0,
 	    "Timestamp file `%s' is owned by uid=%d, but expected owner=%d.\n\
@@ -475,6 +613,56 @@ char *cmd, *user, *salt, *encr;
     return 1;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Looks up a group name or number (as a text string), returns gid.
+ * Accepts special names (e.g. <caller> or <owner>).
+ * Returns -1 if no such group.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+int
+findgid(allowbrackets, grouplabel)
+int allowbrackets;	/* accept names like <caller> */
+char *grouplabel;	/* name or numeric form */
+{
+    struct group *gp;
+    SETGRENT_T setgrent();
+    void endgrent();
+    int numeric_gid;
+    int found_gid, is_numeric=0;
+    int l=strlen(grouplabel);
+
+    char c;
+
+    if (allowbrackets && grouplabel[0] == '<' && grouplabel[l-1] == '>') {
+	/* Translate special name */
+	struct passwd *pw = getpwentry(1, grouplabel);
+	if (!pw)
+	    return -1;
+	else
+	    return pw->pw_gid;
+    }
+
+    is_numeric = (sscanf(grouplabel, "%d%c", &numeric_gid, &c) == 1);
+
+    if (is_numeric)
+	return numeric_gid;
+
+    /* Grouplabel didn't look like a number (according to sscanf),
+     * so look up its name.
+     */
+    setgrent();
+    for (found_gid = -1, gp = getgrent(); gp; gp = getgrent()) {
+	if (strcmp(grouplabel, gp->gr_name) == 0) {
+	    /* Found the gid in the group file */
+	    found_gid = gp->gr_gid;
+	    break;
+	}
+    }
+    endgrent();
+
+    return found_gid;
+}
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Adds condition to condition list.
  *	returns -1 on syntax error, malloc error, etc;
@@ -487,19 +675,30 @@ char *condition;	/* condition to insert: xxx~yyy */
 char *s;		/* pts to yyy in condition */
 int isglobal;		/* Is this a per-command or global condition? */
 {
-    int invert = (*condition == '!');
+    char **globlist;
     TimeList *tl;
+    int i;
+    int invert = (*condition == '!');
 
     if ( invert )
 	condition++;
 
+    /* All conditions accept {a,b,c}  or a,b,c  as lists.
+     * Form the globlist and pass along...
+     */
+    /* Do brace globbing */
+    if ((i=globbraces(s, 1, &globlist)) != 0)		/* Local Condition */
+	return Error(0, 0, "%tMissing `%c'.\n", i);
+
     if (STRMATCH3("time", condition, s-1)) {
-	tl = isglobal ? &gi.timeafter : &li.time;
-	if (InsertTimeList(s, tl, isglobal ? "global" : "local", invert) == -1)
+	tl = isglobal ? &globalinfo.timeafter : &localinfo.time;
+	if (InsertTimeList(s, globlist,
+			tl, isglobal ? "global" : "local", invert) == -1)
 	    return -1;
 
     } else if (STRMATCH3("user", condition, s-1)) {
-	if (InsertUserList(s, &li.userpats, &li.origtext, invert) == -1)
+	if (InsertUserList(s, globlist,
+		    &localinfo.userpats, &localinfo.origtext, invert) == -1)
 	    return -1;
 
     } else {
@@ -517,18 +716,19 @@ int isglobal;		/* Is this a per-command or global condition? */
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-InsertUserList(wd, sl, otl, invert)
+InsertUserList(wd, wdlist, sl, otl, invert)
 char *wd;	/* Pattern to match; must NOT have leading '!'; braces ok;
 		 * disallow '<', '>'.
 		 */
-Simple2List *sl;/* Insert user list elements at sl->next */
-SimpleList *otl;/* Insert original text to this list, at otl->next */
+char **wdlist;	/* brace-expanded u/g/h list */
+Simple2List *sl;/* Insert user list elements (i.e. argument wdlist)
+		 * at sl->next.
+		 */
+SimpleList *otl;/* Insert original text (i.e. argument wd) * at otl->next.  */
 int invert;	/* Inverts the test */
 {
-    int i, iwd;
+    int iwd;
     char *tok, *s;
-    char **wdlist;
-    int globbraces();
     SimpleList *new;
     Simple2List *new2;
 
@@ -555,10 +755,6 @@ int invert;	/* Inverts the test */
     strcpy(new->pat, wd);
     otl->next = new;
 
-    /* Do brace globbing */
-    if ((i=globbraces(wd, &wdlist)) != 0)
-	return Error(0, 0, "%tMissing `%c'.\n", i);
-
     for (iwd=0; (tok=wdlist[iwd]); iwd++) {
 	new2 = (Simple2List *) malloc(sizeof(Simple2List));
 	if (!new2)
@@ -593,8 +789,8 @@ int isglobal;	/* !0 means its from a global def */
      * Sets matches.match_user if user matches; sets it to 1 if the last
      * match is non-inverting, 0 otherwise.
 
-     * BUT! the time list created by the InsertUserList function
-     * is in reverse order, so we only need to find if the first
+     * BUT! the list created by the InsertUserList function
+     * is in reverse order, so we only need to find the first
      * entry in the list that is a match (+ or -) and stop there!
      */
 
@@ -606,16 +802,16 @@ int isglobal;	/* !0 means its from a global def */
 	if (check_ugh(sl->other->pat, invert ? sl->pat+1 : sl->pat) == 0) {
 	    match = 1;
 	    matches.user = invert ? 0 : 1;
-	    if (debug)
+	    if (debug || it_came_from_cmdline)
 		(void) fprintf(stderr,
 			"\tPermission %s: %s pattern %suser~%s\n",
 			invert ? "denied" : "allowed",
 			isglobal ? "global" : "per-cmd",
-			invert ? "!" : "", sl->pat);
-	} else if (debug) {
+			invert ? "!" : "", sl->other->pat);
+	} else if (debug || it_came_from_cmdline) {
 	    (void) fprintf(stderr, "\tNot applicable: %s pattern %suser~%s\n",
 			isglobal ? "global" : "per-cmd",
-			invert ? "!" : "", sl->pat);
+			invert ? "!" : "", sl->other->pat);
 	}
     }
 }
@@ -658,41 +854,52 @@ char *pat, *host;
 /* Try to match a string to a pattern. */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-match_pattern(match, str, pattern)
+match_pattern(match, do_glob, str, pattern)
 int match;	/* Return input value of match on failure; 1 on success */
+int do_glob;	/* 0: no brace globbing;
+		 * 1: brace glob;
+		 * >1: wrap in braces, then brace-glob.
+		 */
 char *str;
 char *pattern;
 {
-    int i, ipat;
+    int i, ipat, wrap;
     char *tok, tokbuf[1000];
-    char **patlist;
-    int globbraces();
+    char **patlist, *pat1list[2];
     char chkbuf[1024];
 
-    /* Do brace globbing on the pattern */
-    /* fprintf(stderr, "globbraces -> ``%s''\n", pattern); */
-    if ((i=globbraces(pattern, &patlist)) != 0) {
-	Error(0, 0, "%tMissing `%c'.\n", i);
+
+    if (do_glob != 0) {
+	/* Do brace globbing on the pattern */
+	wrap = (do_glob > 1) ? 1 : 0;
+	if ((i=globbraces(pattern, wrap, &patlist)) != 0) {	/* MatchPat */
+	    Error(0, 0, "%tMissing `%c'.\n", i);
+	    return match;
+	}
     } else {
-	for (ipat=0; (tok=patlist[ipat]); ipat++) {
-	    strcpy(tokbuf, tok);
-	    anchor(tok, chkbuf);			/* Anchor all matches */
-	    if ((*pat_compile)(chkbuf) != NULL) {
-		return Error(0, 0,
-			"%t\n\tBad command pattern: `%s'.\n", pattern);
+	pat1list[0] = pattern;
+	pat1list[1] = NULL;
+	patlist = pat1list;
+    }
 
-	    } else if ((*pat_compare)(str) == 1) {
-		    if (debug)
-			(void) fprintf(stderr,
-			    "\tMatched user's command=%s to CmdPattern=%s\n",
-			    str, pattern);
-		    return 1;
+    for (ipat=0; (tok=patlist[ipat]); ipat++) {
+	strcpy(tokbuf, tok);
+	anchor(tok, chkbuf);			/* Anchor all matches */
+	if ((*pat_compile)(chkbuf) != NULL) {
+	    Error(0, 0, "%t\n\tBad command pattern: `%s'.\n", pattern);
+	    return match;
 
-	    } else if (debug) {
-		(void) fprintf(stderr,
-			"\tNo match user's command=%s to CmdPattern=%s\n",
-			    str, pattern);
-	    }
+	} else if ((*pat_compare)(str) == 1) {
+		if (debug)
+		    (void) fprintf(stderr,
+			"\tMatched user's command=%s to CmdPattern=%s\n",
+			str, pattern);
+		return 1;
+
+	} else if (debug) {
+	    (void) fprintf(stderr,
+		    "\tNo match user's command=%s to CmdPattern=%s\n",
+			str, pattern);
 	}
     }
     return match;
@@ -710,9 +917,15 @@ char *token;		/* user/group/host pattern */
 {
     char chkbuf[1024];
     char *userpat, *grouppat, *hostpat;
+    char *colon;
     int match, i;
 
-    if (ui.orig_uid == 0)
+    if (strlen(token) > sizeof(chkbuf)-4) {
+	return Error(0, 1, "%tCan't handle patterns larger than %d chars.\n",
+		sizeof(chkbuf)-4);
+    }
+
+    if (userinfo.caller.pw_uid == 0)
 	return 0;			/* root is always legit */
 
     /* Split into user:group@host; check host part first (if it exists) */
@@ -724,15 +937,15 @@ char *token;		/* user/group/host pattern */
 	match = -1;
 	if (hostpat[0] == '+') {
 
-#ifdef USE_NETGROUP
+#ifdef HAVE_INNETGR
 	    if (hostpat[1] == 0)
 		return Error(0, 0,
 			"%tMissing netgroupname in pattern `%s'.\n", origtext);
-	    match = check_host(hostpat, ui.hostname);
+	    match = check_host(hostpat, userinfo.hostname);
 #else
 		return Error(0, 0, 
 	"%thostnames may not begin with `+' since this super() was compiled\n\
-without -DUSE_NETGROUP.\n");
+without -DHAVE_INNETGR.\n");
 #endif
 	} else {
 	    strtolower(hostpat);
@@ -742,13 +955,19 @@ without -DUSE_NETGROUP.\n");
 	}
  
 	if (match == -1)
-	    match = check_host(hostpat, ui.lc_hostname);
+	    match = check_host(hostpat, userinfo.lc_hostname);
+
+	if (debug > 1)
+	    fprintf(stderr,
+		"\thost pattern <%s> %s user's host <%s>\n",
+		hostpat, (match == -1) ? "did not match" : "matched",
+		userinfo.lc_hostname);
 
 	if (match == -1)
 	    return -1;
     }
 
-    grouppat = strchr(token, ':');
+    colon = grouppat = strchr(token, ':');
     userpat = token;
     if (*token == '\0' && !hostpat) {
 	/* Nothing in pattern?! */
@@ -771,7 +990,7 @@ without -DUSE_NETGROUP.\n");
 	return Error(0, 0, "%t\n\tGroup pattern `%s' contains a colon!\n",
 			grouppat);
 
-    if (gi.group_slash == 0 && strchr(grouppat, '/'))
+    if (globalinfo.group_slash == 0 && strchr(grouppat, '/'))
 		return Error(0, 0, "%t\n\tFormat error in super.tab file: \
 group pattern `%s' contains a slash.\n\
 \tPerhaps you meant to use Cmd::Filename, but forgot one colon,\n\
@@ -779,19 +998,19 @@ group pattern `%s' contains a slash.\n\
 \tslashes in group patterns, use global option group_slash=y.\n", grouppat);
 
 
-#ifdef USE_NETGROUP
+#ifdef HAVE_INNETGR
     if (userpat[0] == '+') {
 	if (userpat[1] == 0)
 	    return Error(0, 0,
 			"%tMissing netgroupname in pattern `%s'.\n", origtext);
-	match = netgrp_u_compare(&userpat[1], ui.caller);
+	match = netgrp_u_compare(&userpat[1], userinfo.caller.pw_name);
     } else
 #endif
     {
 	anchor(userpat, chkbuf);		/* Anchor all matches */
 	if ((*pat_compile)(chkbuf) != NULL)
 	    return Error(0, 0, "%t\n\tbad user pattern: `%s'.\n", origtext);
-	match = (*pat_compare)(ui.caller);
+	match = (*pat_compare)(userinfo.caller.pw_name);
 #ifdef MATCH_DECIMAL_UID
 	if (match != 1) {
 	    /* Enabling MATCH_DECIMAL_UID allows the userpat to be
@@ -801,19 +1020,31 @@ group pattern `%s' contains a slash.\n\
 	     * pattern userpat.
 	     */
 	    char buf[20];
-	    (void) sprintf(buf, "%d", ui.orig_uid);
+	    (void) sprintf(buf, "%d", userinfo.caller.pw_uid);
 	    match = (*pat_compare)(buf);
 	}
 #endif
     }
+    if (debug > 1)
+	fprintf(stderr,
+	    "\tuser pattern <%s> %s username <%s>\n",
+	    userpat, (match != 1) ? "did not match" : "matched",
+	    userinfo.caller.pw_name);
+
     if (match != 1)
 	return -1;
 
     anchor(grouppat, chkbuf);
-    i = ingroup(ui.caller, ui.orig_gid, chkbuf);
+    i = ingroup(userinfo.caller.pw_name, userinfo.caller.pw_gid, chkbuf);
     if (i == -1)
 	return Error(0, 0, "%t\n\tbad group pattern\n", origtext);
-    else if (i != 1)
+
+    if (debug > 1)
+	fprintf(stderr,
+	    "\tuser <%s> is %sa member of group <%s>\n",
+	    userinfo.caller.pw_name, (i != 1) ? "not " : "", grouppat);
+
+    if (i != 1)
 	return -1;
     
     return 0;				/* Success! */
@@ -825,7 +1056,7 @@ group pattern `%s' contains a slash.\n\
 int
 ingroup(user, gid, gp_pat)
 char *user;
-GID_T gid;
+gid_t gid;
 char *gp_pat;	/* pattern to match */
 {
     /* Use:
