@@ -1,7 +1,7 @@
 /* The code should compile with either ANSI C or K&R compilers. */
 
 /*
- *      Copyright (c) 1995 by William Deich.
+ *      Copyright (c) 1995, 1996 by William Deich.
  *      Written by William Deich.  Not derived from licensed software.
 
  *    You may distribute under the terms of either the GNU General Public
@@ -12,170 +12,24 @@
  * prototypes in K&R C: declare functions as:
  *	func_type  funcname __P((arglist));
  */
-#ifdef __STDC__
-#define __P(x) x
-#else
-#define __P(x) ()
-#endif
-
-#ifdef __STDC__
-#ifndef HAVE_LOCALE
-#define HAVE_LOCALE
-#endif
-#ifndef HAVE_LOCALTIME
-#define HAVE_LOCALTIME
-#endif
-#endif
-
-#ifdef SUNOS5_4
-#ifndef SUNOS5
-#define SUNOS5
-#endif
-#endif
-
-#ifdef SUNOS5
-#ifndef SVR4
-#define SVR4
-#endif
-#endif
-
-
-/* Don't define _XOPEN_SOURCE for SCO: although it uses this for compliance
- * reasons, this define causes some problems with password file negotiation.
- *
- * Also don't define it for Solaris 2.4 (SunOS 5.4) and higher: it causes
- * problems for things like struct timespec, which doesn't get defined if
- * _XOPEN_SOURCE is set (instead struct _timespec is defined), but stat.h
- * uses struct timespec in its definition of itimerval.  It's a mess...
- */
-#ifndef _SVR4_SOURCE
-#ifndef SCO
-#ifndef SUNOS5_4
-#define  _XOPEN_SOURCE
-#endif
-#endif
-#endif
-
-#ifdef Digital_UNIX
-#define _OSF_SOURCE
-#include <standards.h>
-#include <signal.h>
-#endif
-
-#include <stdio.h>
-#include <ctype.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <string.h>
-#include <sys/types.h>
-
-#ifdef __STDC__
-#include <locale.h>
-#else
-#ifdef HAVE_LOCALE
-#include <locale.h>
-#endif
-#endif
-
-#ifdef __clix__
-#include <sys/signal.h>
-#ifdef _NSIG
-#define NSIG _NSIG
-#endif
-#else
-#include <signal.h>
-#endif
-
-#include <sys/wait.h>
-#include <sys/param.h>
-
-#include <sys/stat.h>
-#ifndef S_IWGRP
-#define S_IWGRP 0000020
-#endif
-#ifndef S_IWOTH
-#define S_IWOTH 0000002
-#endif
-
-#include <pwd.h>
-#include <grp.h>
-#include <unistd.h>
-
-#ifdef SCO
-#include <limits.h>
-#include <sys/signal.h>
-#include <shadow.h>
-#include <malloc.h>
-#endif
-
-#ifdef _SVR4_SOURCE
-#undef NSIG
-#define NSIG _sys_nsig
-#endif
-
-#ifdef __STDC__
-#include <stdlib.h>
-#else
-#include <malloc.h>
-#endif
-
-#ifdef SVR4
-#include <shadow.h>
-#include <sys/time.h>
-#include <sys/systeminfo.h>
-extern int sysinfo();
-#define gethostname(buf, lbuf) (sysinfo(SI_HOSTNAME, (buf), (lbuf)))
-#undef NSIG
-#define NSIG _sys_nsig
-#else
-#ifndef Digital_UNIX
-extern int gethostname __P((char *, size_t size));
-#endif
-#include <time.h>
-#endif
-
-#ifdef USE_GETHOSTBYNAME
-#include <netdb.h>
-#endif
-
-#ifdef USE_SYSLOG
-#include <syslog.h>
-#endif
-
-
-#ifdef _POSIX_PATH_MAX
-#undef MAXPATHLEN
-#define MAXPATHLEN _POSIX_PATH_MAX
-#else
-#ifndef MAXPATHLEN
-#define MAXPATHLEN 255
-#endif
-#endif
-
-#ifdef __STDC__
-    typedef pid_t PID_T;
-    typedef uid_t UID_T;
-    typedef gid_t GID_T;
-#else
-    typedef int PID_T;
-    typedef int UID_T;
-    typedef int GID_T;
-#endif
-
-/* type returned by setgrent() */
-#ifdef Digital_UNIX
-typedef int SETGRENT_T;
-#else
-typedef void SETGRENT_T;
-#endif
 
 /* ==================================================================== */
 
+#include "localsys.h"
 #include "hsearch.h"
 #include "version.h"
 
 /* ==================================================================== */
 
+#ifndef __STDC__
+    extern char *malloc();
+    extern char *realloc();
+#endif
+
+#ifndef HAVE_ALNUM
+#define alnum(c) (isalpha(c) || isdigit(c))
+#endif
+
 #define SEP	" \t\v\n"		/* How to split fields on input lines */
 #define QM	"\"'"			/* Quotemarks in fields */
 #define CM	"#"			/* Comments in input file */
@@ -185,8 +39,11 @@ typedef void SETGRENT_T;
 
 #define CLEAR_SETTINGS NULL
 
+/* The safe path should normally be defined/edited in config.h.  This is
+ * a just-in-case-it's-missing definition.
+ */
 #ifndef SAFE_PATH
-#define SAFE_PATH "PATH=/bin:/usr/bin:/usr/ucb"
+#define SAFE_PATH "/bin:/usr/bin:/usr/ucb"
 #endif
 
 /* The name under this program assumes it is installed.  If argv[0] isn't
@@ -200,13 +57,14 @@ typedef void SETGRENT_T;
 #define HELP_BASIC	0	/* Basic help shows what you can execute */
 #define HELP_FULL	1	/* Full help on each command */
 #define HELP_FACTS	2	/* Just-the-facts-ma'm mode */
+#define HELP_USAGE	3	/* Print usage only, no per-cmd help */
 
 #ifndef SUPERFILE
 #define SUPERFILE "/usr/local/lib/super.tab"
 #endif
 
-#ifndef TIMESTAMP_DIR
-#define TIMESTAMP_DIR "/usr/local/lib/superstamps"
+#ifndef PERUSER_SUPERFILE
+#define PERUSER_SUPERFILE ".supertab"
 #endif
 
 #ifndef MAXFD
@@ -225,6 +83,11 @@ int getdtablesize __P(( void ));
  */
 #define MAXSETENV 40
 
+/* Default value for the maximum length of a user-supplied environment
+ * variable definition.
+ */
+#define MAXENVLEN 1000
+
 /* If defined, then user patterns are allowed to match the uid as well as
  * the actual username.  We DISABLE this by default because users are
  * almost always identified by username.
@@ -241,20 +104,92 @@ int getdtablesize __P(( void ));
 /* maximum number of tries at entering the password */
 #define MAXTRY 3
 
+/* default maxlength per arg, maxlength all args */
+#define MAXLEN1ARG 1000
+#define MAXLENARGS 10000
 
-#ifdef USE_NETGROUP
+#ifdef HAVE_INNETGR
 extern int innetgr();
 #endif
 
-#ifdef NEED_MEMSET
+#ifndef HAVE_MEMSET
 void *memset __P(( void *s, int c, size_t n ));
 #endif
 
-extern char *prog;			/* this program */
-extern int debug;
-extern int check_syntax;		/* just check syntax of superfile */
+/* special code to indicate we haven't specified a uid yet */
+#define UID_NOTSET ((uid_t) ~0)
 
-extern char *superfile;			/* The actual superfile to be opened. */
+/* special code to indicate we haven't specified a gid yet */
+#define GID_NOTSET ((gid_t) ~0)
+
+/* special code to indicate we haven't specified supplementary groups */
+#define GROUPS_NOTSET ((GETGROUPS_T) ~0)
+
+
+/* Number of elements in an array */
+#define NELEM(x) (sizeof(x)/(sizeof(*x)))
+
+/* n rounded up to a multiple of m -- both should be positive */
+#define ROUNDUP(n, m) ( ((n) % (m) == 0) ? (n) : ((n)/(m) + 1)*(m) )
+
+/* STRBEG evaluates to !0 if s1 begins with substring s2 */
+#define STRBEG(s1, s2) (strncmp((s1), (s2), strlen(s2)) == 0)
+ 
+/* STRMATCH3 expects 'end' to be a pointer into s2, and evaluates to !0
+ * if characters s2..(end-1) fully match s1 (that is, the next character in
+ * s1 is null.
+ */
+#define STRMATCH3(s1, s2, end) \
+	       (strncmp(s1, s2, end-s2) == 0 && s1[end-s2]=='\0')
+		    
+
+#define UNTIL(e) while(!(e))
+
+/* ========================================================================= */
+
+/*
+ * Super's main external variables
+ */
+    extern char *prog;			/* this program */
+    extern int debug;			/* debug level; 0=off */
+    extern int check_syntax;		/* just check syntax of superfile */
+    extern char *superfile;		/* The actual superfile to be opened. */
+    extern int it_came_from_cmdline;	/* Set by -F/-T/-U/-G flags */
+
+/* The following external variables control the Error() routine.
+ * They are modified at various points by super(), to give detailed control
+ * over the error messages that are printed.
+ */
+    extern int error_stderr;		/* stderr() bool */
+    extern int error_syslog;		/* syslog() bool */
+    extern char *error_rlog_host;	/* where rsyslog() msgs are sent */
+    extern char *error_command;		/* our program name */
+    extern char *error_user;		/* our caller's username */
+
+    extern int error_line;		/* input line number of error */
+    extern int error_nl;		/* number of lines this msg refers to */
+    extern char *error_srcfile;		/* filename of input with error */
+
+/*
+ * For use with strqtokS -- see that routine for definitions of strqS_xxx
+ */
+    extern unsigned char *strqS_qm;	/* For fast access by strqtokS */
+    extern unsigned char my_qm[256];
+    extern unsigned char *strqS_cc;	/* For fast access by strqtokS */
+    extern unsigned char my_cc[256];
+ 
+
+/*
+ * Extern variable and routines used to compile/match user/group/host patterns.
+ */
+extern char *(*pat_compile) __P(( char *));
+extern int (*pat_compare) __P(( char *));
+extern int need_re_anchor;
+/* ========================================================================= */
+
+/*
+ * Basic structures from which we construct other, bigger entities.
+ */
 
 struct simpleList {
     char *pat;
@@ -321,6 +256,15 @@ struct conditions {
 };
 typedef struct conditions Conditions;
 
+/* The progList struct is for listing all Cmd::File pairs on a control line. */
+struct progList {
+    char *Cmd;
+    char *File;
+};
+typedef struct progList ProgList;
+
+/* ========================================================================= */
+
 /* Password information */
 struct passInfo {
     int required;	/* -1 means not in use */
@@ -336,17 +280,13 @@ struct logInfo {
     FILE *fp;			/* logfile pointer */
     char filename[1024];	/* logfile name */
     char user[1024];		/* value of loguid=xxx, if entered */
-    UID_T uid;			/* UID under which we open logfile */
+    uid_t uid;			/* UID under which we open logfile */
+    pid_t pid;			/* PID of the logger; -1 means none running. */
+    unsigned char newfile;	/* Bool: !0 if logfile given but not yet used */
+    unsigned char newuid;	/* Bool: !0 if loguid given, but not yet used */
 };
 typedef struct logInfo LogInfo;
 
-/* The progList struct is for listing all Cmd::File pairs on a control line. */
-struct progList {
-    char *Cmd;
-    char *File;
-};
-typedef struct progList ProgList;
-
 /* progMatch is for keeping track of matches in a ProgList */
 struct progMatch {
     ProgList *cmd_file;
@@ -362,11 +302,16 @@ struct progMatch {
 };
 typedef struct progMatch ProgMatch;
 
-/* Global information from the :global lines */
+/* ========================================================================= */
+
+/*
+ * Global information from the :global lines
+ */
 struct globalInfo {
-    char owner[32];	/* File owner must be this person; overridden
+    char owner[32];	/* Owner of FullPath must be this person; overridden
 			 * by local owner=xxx option, if present.
 			 */
+    char *chdir_path;	/* Change to this dir before exec'ing; null if none */
     int relative_path;	/* Allow filenames to be relative.  This is
 			 * in general a stupid idea.  Don't do it!
 			 */
@@ -374,10 +319,16 @@ struct globalInfo {
 			 * allow this, you make it harder to catch certain
 			 * command-line typos.  Don't do it!
 			 */
+    int maxenvlen;	/* max length of envvar (all of "name=value") */
+    char **env;		/* null-terminated list of vars from env=var[,...] */
+
     int nice_incr;	/* value of the nice=nnn increment value */
 
     int mask;		/* umask setting */
 
+    long maxlen1arg;	/* max len of any single arg */
+    long maxlenargs;	/* max len of all args, combined */
+
     PassInfo passinfo;	/* password information */
 
     Simple2List userbefore;	/* list of u/g/h pats before per-cmd pats */
@@ -394,21 +345,28 @@ struct globalInfo {
     LogInfo log;		/* Information for logging to file */
 
     char mailcmd[500];		/* Information for logging via mail */
-    int mail_success;	/* bool: mail on successful tries? (-1 = unknown) */
+    int mail_success;		/* bool: mail on success? (-1 = unknown) */
+    int gethostbyname;		/* bool: use gethostbyname()? */
+
+    GETGROUPS_T groups[NGROUPS_MAX];	/* groups from [add]groups=xxx,... */
+    int ngroups;		/* number of supplementary groups */
+    int groups_added;		/* bool: groups were addgroups=, not groups= */
 };
 typedef struct globalInfo GlobalInfo;
 
-/* A structure that holds information describing the caller */
+/*
+ * Information describing the caller
+ */
 struct userInfo {
-    char caller[128];			/* who's invoking prog */
+    struct passwd caller;		/* who's invoking program */
     char hostname[MAXHOSTNAMELEN];	/* whence came the user */
     char lc_hostname[MAXHOSTNAMELEN];	/* hostname in lower case */
-    UID_T orig_uid;			/* caller's original uid */
-    GID_T orig_gid;			/* caller's original gid */
     int orig_mask;			/* umask setting at startup */
 
-    UID_T new_uid;			/* new uid, from uid=xxx or u+g=xxx */
-    GID_T new_gid;			/* new gid, from gid=xxx or u+g=xxx */
+    uid_t orig_uid;			/* uid at prog start, from getuid() */
+    gid_t orig_gid;			/* gid at prog start, from getgid() */
+    uid_t new_uid;			/* new uid, from uid=xxx or u+g=xxx */
+    gid_t new_gid;			/* new gid, from gid=xxx or u+g=xxx */
 
     OurTime ourtime;			/* when we started, etc */
 
@@ -417,19 +375,29 @@ struct userInfo {
 };
 typedef struct userInfo UserInfo;
 
-/* A structure that holds the per-cmd information */
+/*
+ * Per-entry (in the super.tab file) information.  This gets filled in
+ * at various points, as the program learns more.
+ */
 struct localInfo {
     ProgMatch progs;	/* Records prog::file sets, and is updated w/ matches */
 
     char *info;		/* value of info=xxx option */
     char *die;		/* Gets msg from die=msg ; null if none */
     char *print;	/* Gets msg from print=msg ; null if none */
+    char *chdir_path;	/* Change to this dir before exec'ing; null if none */
 
     char user[32];	/* value of uid=xxx option */
     char group[32];	/* value of gid=xxx option */
     char u_g[32];	/* value of u+g=xxx option */
     char owner[32];	/* value of owner=xxx option */
-    char *env;		/* value of env=var[,...] option */
+    uid_t file_uid;	/* uid of the matched FullPath */
+    gid_t file_gid;	/* gid of the matched FullPath */
+    GETGROUPS_T groups[NGROUPS_MAX];	/* groups from [add]groups=xxx,... */
+    int ngroups;	/* number of supplementary groups */
+    int groups_added;	/* bool: groups were addgroups=, not groups= */
+    int maxenvlen;	/* max length of envvar (all of "name=value") */
+    char **env;		/* null-terminated list of vars from env=var[,...] */
     char *setenv[MAXSETENV+1];	/* values of setenv=var[,...] option */
     char *fdlist;	/* value of fd=nnn[,...] option */
     int mask;		/* value of umask=xxx option */
@@ -438,6 +406,8 @@ struct localInfo {
     SimpleList origtext;	/* list of PermittedUser patterns */
     TimeList time;	/* A list of the actual time ranges used */
     int usr_args[2];	/* number of user-entered args allowed */
+    long maxlen1arg;	/* max len of any single arg */
+    long maxlenargs;	/* max len of all args, combined */
     StrArray argpats;	/* argNNN=xxx arguments */
 
     LogInfo log;	/* Information for logging to file */
@@ -450,161 +420,132 @@ struct localInfo {
 };
 typedef struct localInfo LocalInfo;
 
+/* ========================================================================= */
+
 extern FileList *currfile;		/* list of currently-open files */
 
-extern GlobalInfo gi;			/* :global info */
+extern GlobalInfo globalinfo;		/* :global info */
 
-extern UserInfo ui;			/* User's info */
+extern UserInfo userinfo;		/* User's info */
 
-extern LocalInfo li;			/* per-cmd info */
+extern LocalInfo localinfo;		/* per-cmd info */
 
 extern Conditions matches;		/* To keep track of what matched */
 
-extern int error_stderr;		/* stderr() bool, used by Error() */
-extern int error_syslog;		/* syslog() bool, used by Error() */
-extern char *error_command;		/* command used by Error() */
-extern char *error_user;		/* username printed by Error() */
-
-extern int error_line;			/* line number, used by Error() */
-extern int error_nl;			/* number of lines, used by Error() */
-extern char *error_srcfile;		/* filename, used by Error() */
-
-/* For use with strqtokS */
-extern unsigned char *strqS_qm;		/* For fast access by strqtokS */
-extern unsigned char my_qm[256];
-extern unsigned char *strqS_cc;		/* For fast access by strqtokS */
-extern unsigned char my_cc[256];
- 
-char	*strqtokS __P(( char *s, char *delim,
-		char *quotemarks, char *commentchars, unsigned int flags));
-void	init_strqtokS __P(( void ));
-
-int	do_options(int argc, char **argv, int *help, int *vers, int *verbose);
-int	get_canonical_hostname __P((char *buf, int len));
-int	basic_userinfo __P( (void) );
-char	*approve __P((char *usrcmd, int help, int version, int verbose));
-int	check_owner __P(( char *file ));
-char	**newargs __P((char *path_plus, char **argv, int *n_builtin));
-char	**buttonup __P((char *));
-int	set_u_g __P(( void ));
+/* ========================================================================= */
+void	add_builtin_variables __P(( void ));
+#ifdef HAVE_SYSINFO
+void	add_sysinfo_variables __P(( void ));
+#endif
+#ifdef HAVE_UNAME
+void	add_uname_variables __P(( void ));
+#endif
+int	add_variable __P(( char *varname, char *vardefn ));
+void	anchor __P(( char *in, char *out ));
+char*	approve __P((char *usrcmd, int givehelp, int verbose));
+int	blkfree __P((char **));
+char ** blkdup __P((char **blk));
+int	build_cmd_file __P((char *, int , char *, char **));
+char**	buttonup __P((char *));
+int	check_arglistlen __P(( char **argv ));
+int	check_owner __P(( void ));
 int	check_pass __P(( char *cmd ));
 int	checkenv __P(( char *name,  char *value,  char *pat ));
-
-int	parseline __P((int givehelp, int checksyntax, char *buf, char *usrcmd));
-void	match_ugh_user __P((Simple2List *sl, int isglobal));
-
-char	*str_val __P(( char *left, int sep, char *str ));
-void	strtolower __P(( char *string ));
-char	*ends __P(( char *s1, char *s2 ));
-char	*clean_buf __P(( char *buf, char *outbuf ));
-char	*fillbuffer __P(( FILE *fp, int *indentok, int *nl ));
-
-int	Error __P(( int show_perror, int die, char * fmt, ... ));
-
-int	get_password __P(( char *cmd, char *user,  char *salt,  char *encr ));
-int	get_encrypted_pw __P(( void ));
-char	*Getenv __P(( char *s ));
-int	getlogdir __P((char *user, char *buf));
-int	ingroup __P(( char * user, GID_T gid, char * gp_pat ));
-void	logmsg __P(( char * cmd, char ** args ));
-
-int	makedir __P(( char *directories ));
-char	*makedirname __P(( char *prefix, char *hostname, char *path ));
-
-FILE	*open_writer __P(( char *user, char *filename ));
+char*	clean_buf __P(( char *buf, char *outbuf ));
 void	close_writer __P(( void ));
-void	opensuperlog __P(( void ));
-void	printhelp __P(( int verbose ));
-void	printhelp_hello __P(( int verbose, char *usrcmd ));
-
-void	re_anchor __P(( char *in, char *out));
-int	re_exec __P(( char *));
-char	*re_comp __P(( char *));
-void	anchor __P(( char *in, char *out ));
-int	shell_compare __P(( char *));
-char	*shell_compile __P(( char *));
-int	match_pattern __P((int match, char *str, char *pattern));
-
-void	StrInit __P(( StrArray *a ));
-int	StrNelts __P(( StrArray *a ));
-int	StrLastInUse __P(( StrArray *a ));
-char	*StrEltGetPtr __P(( StrArray *a, int ielt ));
-int	StrEltCpy __P(( StrArray *a, int ielt, char *str));
-void	StrEltsUnused __P(( StrArray *a ));
-int	StrNalloc __P(( StrArray *a, int nelt));
-
-int	stringcopy __P(( char *to, char *from, int n));
-
-char	*dayname __P(( int daynum ));
+int	colon_define __P(( char *command ));
+int	colon_global __P(( char *command ));
+int	colon_include __P(( char *command, int allow_missing ));
+int	colon_if __P(( char *command, int *colonif_succeeded ));
+int	conditions_and_options __P((char *cond_or_opt_wd));
+char*	dayname __P(( int daynum ));
 int	daynum __P(( int unixtime ));
-
-int	InsertCondition __P(( char *condition, char *s, int isglobal ));
-int	InsertTimeList __P(( char *str, TimeList *tl, char *type, int invert));
-void	matchtime __P(( OurTime *our, TimeList *tl ));
-void	readtime_init __P(( void ));
-int	readtime __P(( char *str, short *t1, short *t2, char *d ));
-
-FileList *file_open __P(( FileList *list, char *name ));
-FileList *file_close __P(( FileList *list ));
-
-void	debug_print __P((char *path, char **arglist, int n_builtin));
+void	debug_print __P((char *path, char **arglist, char **envp, int n_bltin));
 void	debug_hello __P((void));
-
-void	init_umask __P((int is_global));
-void	store_umask __P((int mask, int is_global));
-void	set_umask __P((void));
-int	rcl_umask __P((void));
-
-void	init_nice_incr __P((int is_global));
-void	store_nice_incr __P((int nice_incr, int is_global));
-int	set_nice_incr __P((void));
-int	rcl_nice_incr __P((void));
-
-int	InsertUserList __P((char *, Simple2List *, SimpleList *, int ));
+int	do_options __P((int argc, char **argv, int *help, int *vers,
+					int *verbose, int *printvars));
+char*	do_variables __P(( char *str));
+void	print_variable __P(( int indx, char *key, char *data ));
+char*	ends __P(( char *s1, char *s2 ));
+#ifdef HAVE_STDARG_H
+int	Error __P(( int show_perror, int die, char * fmt, ... ));
+#else
+int	Error();
+#endif
+FileList* file_open __P(( FileList *list, char *name, int allow_missing ));
+FileList* file_close __P(( FileList *list ));
+char*	fillbuffer __P(( FILE *fp, int *indentok, int *nl ));
+int	findgid __P(( int allowbrackets, char *grouplabel ));
+int	findgroup __P(( char *grouplabel ));
+int	fixup_fullpath __P((int , char *, char *, char *, int ));
 void	free_SimpleList __P(( SimpleList *));
 void	free_Simple2List __P(( Simple2List *));
 void	free_TimeList __P(( TimeList *));
-
-char	*do_variables __P(( char *str));
-
-int	process_colon_cmds __P(( char *command ));
-int	colon_global __P(( char *command ));
+int	get_canonical_hostname __P((char *buf, int len));
+int	get_encrypted_pw __P(( void ));
+int	get_owner __P((char *path, uid_t *uid, gid_t *gid));
+int	get_password __P(( char *cmd, char *user,  char *salt,  char *encr ));
+char*	Getenv __P(( char *s ));
+int	Getgroups __P((int, GETGROUPS_T *));
+int	getlogdir __P((char *user, char *buf));
+struct passwd *
+	getpwentry __P((int allow_brackets, char *username));
 int	global_arg __P((char *word));
-int	colon_define __P(( char *command ));
-int	add_variable __P(( char *varname, char *vardefn ));
-int	colon_include __P(( char *command ));
-
-int	build_cmd_file __P((char *usrcmd, int givehelp_or_checksyntax,
-					char *word1, char **cond_or_opt_wd));
-int	fixup_fullpath __P((int indx, char *usrcmd, char *path,
-					char *FullPath, int l_FullPath));
-int	conditions_and_options __P((char *cond_or_opt_wd));
+int	globbraces __P(( char *s, int wrap_in_braces, char ***globlist ));
 int	handle_option __P((char *word, char *s, int isglobal));
+int	ingroup __P(( char * user, gid_t gid, char * gp_pat ));
+void	init_nice_incr __P((int is_global));
+void	init_strqtokS __P(( void ));
+void	init_umask __P((int is_global));
+void	init_globalinfo __P( (void) );
+void	init_localinfo __P( (void) );
+int	init_userinfo __P( (void) );
+int	InsertCondition __P(( char *, char *, int ));
+int	InsertTimeList __P(( char *, char **, TimeList *, char *, int ));
+int	InsertUserList __P((char *, char **, Simple2List *, SimpleList *, int));
+void	logmsg __P(( char * cmd, char ** args ));
+int	makedir __P(( char *directories ));
+char*	makedirname __P(( char *prefix, char *hostname, char *path ));
+int	match_pattern __P((int match, int glob, char *str, char *pattern));
+void	match_ugh_user __P((Simple2List *sl, int isglobal));
+void	matchtime __P(( OurTime *our, TimeList *tl ));
+char**	newargs __P((char *path_plus, char **argv, int *n_builtin));
+FILE*	open_writer __P(( char *user, char *filename, pid_t *pid_p ));
+void	opensuperlog __P(( void ));
 int	option_clear_settings __P((char *word, char *s, int isglobal));
 int	option_global_reset_settings __P((void));
 int	option_local_clear_settings __P((void));
-
-
-/* Routines used to compile/match user/group/host patterns */
-extern char *(*pat_compile) __P(( char *));
-extern int (*pat_compare) __P(( char *));
-extern int need_re_anchor;
-
-/* Number of elements in an array */
-#define NELEM(x) (sizeof(x)/(sizeof(*x)))
-
-/* n rounded up to a multiple of m -- both should be positive */
-#define ROUNDUP(n, m) ( ((n) % (m) == 0) ? (n) : ((n)/(m) + 1)*(m) )
-
-/* STRBEG evaluates to !0 if s1 begins with substring s2 */
-#define STRBEG(s1, s2) (strncmp((s1), (s2), strlen(s2)) == 0)
- 
-/* STRMATCH3 expects 'end' to be a pointer into s2, and evaluates to !0
- * if characters s2..(end-1) fully match s1 (that is, the next character in
- * s1 is null.
- */
-#define STRMATCH3(s1, s2, end) \
-	       (strncmp(s1, s2, end-s2) == 0 && s1[end-s2]=='\0')
-		    
-
-#define UNTIL(e) while(!(e))
+int	parseline __P((int givehelp, int checksyntax, char *buf, char *usrcmd));
+void	printhelp __P(( int verbose ));
+void	printhelp_hello __P(( int verbose, char *usrcmd ));
+int	process_colon_cmds __P(( char *command, int *colonif_succeeded ));
+int	process_logfile_opt __P(( void ));
+int	rcl_nice_incr __P((void));
+int	rcl_umask __P((void));
+void	re_anchor __P(( char *in, char *out));
+char*	s_re_comp __P(( char *));
+int	s_re_exec __P(( char *));
+void	readtime_init __P(( void ));
+int	readtimerange __P(( char *str, short *t1, short *t2, char *d ));
+int	readtime __P(( char *str, short *t1, char *d ));
+int	set_chdir __P((void));
+int	set_nice_incr __P((void));
+int	set_u_g __P(( void ));
+void	set_umask __P((void));
+int	Setgroups __P((int, GETGROUPS_T *));
+int	shell_compare __P(( char *));
+char*	shell_compile __P(( char *));
+void	store_nice_incr __P((int nice_incr, int is_global));
+void	store_umask __P((int mask, int is_global));
+char*	str_val __P(( char *left, int sep, char *str ));
+int	StrEltCpy __P(( StrArray *a, int ielt, char *str));
+char*	StrEltGetPtr __P(( StrArray *a, int ielt ));
+void	StrEltsUnused __P(( StrArray *a ));
+void	StrInit __P(( StrArray *a ));
+int	stringcopy __P(( char *to, char *from, int n));
+int	StrLastInUse __P(( StrArray *a ));
+int	StrNalloc __P(( StrArray *a, int nelt));
+int	StrNelts __P(( StrArray *a ));
+void	strtolower __P(( char *string ));
+char*	strqtokS __P(( char *, char *, char *, char *, unsigned int ));
+int	user_supertab __P(( char *command ));
