@@ -1,3 +1,4 @@
+static char rcsid[] = "$Id: utils.c,v 1.89 1998/04/07 08:04:06 will Exp $";
 #include "super.h"
 
 /* Variables table flag -- !0 means has been created */
@@ -51,6 +52,7 @@ char *user;
 char *buf;
 {
     /* Gets the login directory of the named user, and puts it into buf.
+     * The "user" argument can either be a username or a uid in text form.
      * If user==NULL || *user == '\0', the current user is obtained.
      * Best if buf is MAXPATHLEN long.
      * 0 is returned on success; -1 on error.
@@ -58,18 +60,14 @@ char *buf;
  
     struct passwd *pass;
     char *p;
-    char *getlogin();
  
     buf[0] = '\0';
     if (user != NULL && *user != '\0') {
-	/* Name given; use getpwnam */
-	pass = getpwnam(user);
-    } else if ((p = getlogin()) != NULL) {
-	/* No name given; use current login name */
-	pass = getpwnam(p);
+	/* Name or uid given; use getpwentry */
+	pass = getpwentry(1, user);
     } else {
-	/* No user given && getlogin() returned NULL; use current uid */
-	pass = getpwuid(getuid());
+	/* No user given; use current uid */
+	pass = getpwuid(userinfo.orig_uid);
     }
  
     if (pass == (struct passwd *) NULL)
@@ -113,15 +111,6 @@ char *out;
     out[i] = '\0';
 }
 
-#ifdef NEED_ALNUM
-int
-alnum(c)
-int c;
-{
-    return isalpha(c) || isdigit(c);
-}
-#endif
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Grow an expandable buffer */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -315,9 +304,10 @@ int *nl;		/* Returned with number of lines read in */
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 FileList *
-file_open(parent, name)
+file_open(parent, name, allow_missing)
 FileList *parent;
 char *name;
+int allow_missing;	/* silently ignore missing files if parent != null. */
 {
     char *s;
 
@@ -325,13 +315,14 @@ char *name;
     struct stat st;
     static int nobody_uid = -999;
     static struct passwd *nobodypw;
+    int uid, euid;
 
     if (nobody_uid == -999) {
 	nobodypw = getpwnam("nobody");
 	nobody_uid = nobodypw ? nobodypw->pw_uid : 65534;
     }
 
-    fl = malloc(sizeof(FileList));
+    fl = (FileList *) malloc(sizeof(FileList));
     if (!fl) {
 	Error(0, 0,
 	"%tFailed to malloc list member space for include file <%s>\n", name);
@@ -377,41 +368,105 @@ char *name;
 	    fprintf(stderr, "\tLine %d: opening include file %s\n",
 		parent->line, fl->fullname);
 
-    /* For any file: make sure that, if we are running as root,
-     * the super.tab file is owned by root or nobody, and is not writable
-     * by group/world.
+    /* For any file:
+     *	If we are running as root:
+     *		make sure that the super.tab file is owned by
+     *		root or nobody, and is not writable by group/world.
+     *	Otherwise, if our real uid differs from that under which we
+     *		started the program, make sure the super file is owned
+     *		by ourselves and not group/world writable.
      */
-    if (geteuid() == 0) {
-	if (stat(fl->fullname, &st) == -1) {
+    if (stat(fl->fullname, &st) == -1) {
+	if (errno == ENOENT && allow_missing) {
+	    if (debug)
+		fprintf(stderr, "\tIgnoring missing file %s\n", fl->fullname);
+	} else {
 	    Error(1, 0, "Couldn't stat super.tab %sfile `%s': ",
-			parent == NULL ? "" : "include ", fl->fullname);
+		    parent == NULL ? "" : "include ", fl->fullname);
+	}
+	if (fl->fullname != fl->givenname)
+	    free(fl->fullname);
+	free(fl->givenname);
+	free(fl);
+	if (errno == ENOENT && allow_missing) {
+	    return parent;
+	} else {
 	    return NULL;
 	}
-
+    }
+    euid = geteuid();
+    uid = userinfo.orig_uid;
+    if (euid == 0 && uid == 0) {
+	/* Being run by root */
 	if (st.st_uid != 0 && st.st_uid != nobody_uid) {
-	    if (getuid() == 0) {
-		Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
-or nobody,\n\but we are being run by root.  Bailing out.\n",
-			parent == NULL ? "" : "include ", fl->fullname);
-		return NULL;
+	    Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
+or nobody,\n\tbut we are being run by root.  Bailing out.\n",
+		parent == NULL ? "" : "include ", fl->fullname);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		  "If this weren't debug mode, super would have quit here.\n");
 	    } else {
-		Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
-or nobody,\n\but our euid==0.   Reverting to uid=%d.\n",
-		parent == NULL ? "" : "include ", fl->fullname, getuid());
-		setuid(getuid());
+		return NULL;
 	    }
 	}
 	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
-	    if (getuid() == 0) {
-		Error(0, 0, "super.tab %sfile `%s' is owned by root \
-or nobody,\n\but is group- or world-writeable.  Bailing out.\n",
-		    parent == NULL ? "" : "include ", fl->fullname);
-		return NULL;
+	    Error(0, 0, "super.tab %sfile `%s' is owned by root \
+or nobody,\n\tbut is group- or world-writeable.  Bailing out.\n",
+		parent == NULL ? "" : "include ", fl->fullname);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		  "If this weren't debug mode, super would have quit here.\n");
 	    } else {
-		Error(0, 0, "super.tab %sfile `%s' is owned by root \
-or nobody,\n\but is group- or world-writeable.  Reverting to uid=%d.\n",
-		    parent == NULL ? "" : "include ", fl->fullname, getuid());
-		setuid(getuid());
+		return NULL;
+	    }
+	}
+
+    } else if (euid == 0) {
+	/* Running as root, original user someone else */
+	if (st.st_uid != 0 && st.st_uid != nobody_uid) {
+	    Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
+or nobody,\n\tbut our euid==0.   %s to uid=%d.\n",
+		parent == NULL ? "" : "include ", fl->fullname,
+		(debug || it_came_from_cmdline) ?
+		    "If this weren't debug mode, we'd revert" : "Reverting",
+		userinfo.orig_uid);
+	    if (!debug)
+		setuid(uid);
+	}
+	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
+	    Error(0, 0, "super.tab %sfile `%s' is owned by root \
+or nobody,\n\tbut is group- or world-writeable.  %s to uid=%d.\n",
+		parent == NULL ? "" : "include ", fl->fullname,
+		(debug || it_came_from_cmdline) ?
+		    "If this weren't debug mode, we'd revert" : "Reverting",
+		userinfo.orig_uid);
+	    if (!debug)
+		setuid(uid);
+	}
+
+    } else if (uid != userinfo.caller.pw_uid) {
+	/* Real uid has changed -- make sure uid is owner of superfile. */
+	if (st.st_uid != uid) {
+	    Error(0, 0, "super.tab %sfile `%s' isn't owned by uid %d.\n",
+		parent == NULL ? "" : "include ", fl->fullname, uid);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		    "If this weren't debug mode, super would quit here.\n");
+	    } else {
+		return NULL;
+	    }
+	}
+	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
+	    if (userinfo.orig_uid == 0) {
+		Error(0, 0,
+	    "super.tab %sfile `%s' is group or world-writable.  Bailing out.\n",
+		    parent == NULL ? "" : "include ", fl->fullname);
+		if (debug || it_came_from_cmdline) {
+		    Error(0, 0,
+			"If this weren't debug mode, super would have quit here.\n");
+		} else {
+		    return NULL;
+		}
 	    }
 	}
     }
@@ -537,7 +592,7 @@ StrArray *a;	/* StrArray with element to fill */
     int i;
     int last;
 
-    for (i=0, last=-1; i < a->n; i++) {
+    for (i=0, last = -1; i < a->n; i++) {
 	if (a->str[i].used)
 	    last = i;
     }
@@ -574,10 +629,11 @@ int ielt;
 	unsigned int new_n;
 	if (a->n == 0) {
 	    new_n = (ielt < 4) ? 4 : 2*ielt ;
-	    a->str = malloc(sizeof(CountedString) * new_n);
+	    a->str = (CountedString *) malloc(sizeof(CountedString) * new_n);
 	} else {
 	    new_n = 2*ielt;
-	    a->str = realloc((void *) a->str, sizeof(CountedString) * new_n);
+	    a->str = (CountedString *)
+		    realloc((void *) a->str, sizeof(CountedString) * new_n);
 	}
 	if (!a->str)
 	    return -1;
@@ -619,17 +675,17 @@ opensuperlog()
 {
     extern FILE *error_logfile;	/* to tell Error() where the log is */
 
-    if (gi.log.fp)
-	(void) fclose(gi.log.fp);
+    close_writer();		/* in case there is already one running */
 
-    if (*gi.log.filename == '\0') {
-	(void) Error(0, 0, "opensuperlog(): logfile name is (nil)\n");
+    if (*globalinfo.log.filename == '\0') {
+	Error(0, 0, "opensuperlog(): logfile name is (nil)\n");
 	return;
     }
 
-    gi.log.fp = open_writer(gi.log.user, gi.log.filename);
+    globalinfo.log.fp = open_writer(globalinfo.log.user,
+				globalinfo.log.filename, &globalinfo.log.pid);
 
-    error_logfile = gi.log.fp;		/* so that Error() writes here too */
+    error_logfile = globalinfo.log.fp;	/* ...so Error() writes here too */
     return;
 }
 
@@ -644,9 +700,10 @@ opensuperlog()
      */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 FILE *
-open_writer(user, filename)
+open_writer(user, filename, pid_p)
 char *user;
 char *filename;
+pid_t *pid_p;	/* filled with pid of created logger */
 {
     FILE *fp = NULL;
     int fildes[2];
@@ -663,6 +720,8 @@ char *filename;
 
     } else if (child > 0) {
 	/* In parent -- close read side, and aim logstream at write side */
+	if (pid_p)
+	    *pid_p = child;
 	(void) close(fildes[0]);
 	if (!(fp = fdopen(fildes[1], "w"))) {
 	    (void) Error(1, 0,
@@ -683,24 +742,24 @@ char *filename;
 	    exit(1);
 	}
 	if (user && *user != '\0') {
-	    stringcopy(li.user, user, sizeof(li.user));
-	    *li.group = '\0';
-	    *li.u_g = '\0';
+	    stringcopy(localinfo.user, user, sizeof(localinfo.user));
+	    *localinfo.group = '\0';
+	    *localinfo.u_g = '\0';
 	    if (set_u_g() == -1) {
 		(void) Error(1, 0,
 		"failed to setuid %s before opening logfile; no logging: ",
-		gi.log.user);
+		globalinfo.log.user);
 		exit(1);
 	    }
 	}
-	if (!(fp = fopen(gi.log.filename, "a"))) {
+	if (!(fp = fopen(globalinfo.log.filename, "a"))) {
 	    if (user && *user != '\0')
 		(void) Error(1, 0,
 		    "failed to open logfile `%s' using uid `%s': ",
-		    gi.log.filename, li.user);
+		    globalinfo.log.filename, localinfo.user);
 	    else
 		(void) Error(1, 0,
-			"failed to open logfile `%s': ", gi.log.filename);
+		    "failed to open logfile `%s': ", globalinfo.log.filename);
 	    exit(1);
 	}
 	while (fgets(text, sizeof(text), input)) {
@@ -720,12 +779,27 @@ char *filename;
 void
 close_writer()
 {
-    if (!gi.log.fp)
+    pid_t pid;
+
+    if (!globalinfo.log.fp && globalinfo.log.pid == -1)
 	return;
 
-    if (fclose(gi.log.fp) == EOF)
-	(void) Error(1, 0, "failed to close gi.log.fp: ");
-    (void) wait((int *) NULL);
+    if (globalinfo.log.fp) {
+	if (fclose(globalinfo.log.fp) == EOF)
+	    Error(1, 0, "failed to close globalinfo.log.fp: ");
+	globalinfo.log.fp = NULL;
+    }
+
+    if (globalinfo.log.pid != -1) {
+	while ((pid = wait((int *) NULL)) > 0 && pid != globalinfo.log.pid) {
+	    Error(0, 0, "wait() surprised! close_writer() received pid %d;\n\t\
+expected logger pid = %d; waiting for correct pid...\n",
+			pid, globalinfo.log.pid);
+	}
+	if (pid == -1)
+	    Error(1, 0, "while waiting for logger process to exit");
+	globalinfo.log.pid = -1;
+    }
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -762,7 +836,7 @@ char *prefix, *hostname, *file;
     if (!*hostname)
 	return file;
 
-#ifdef L14
+#ifndef HAVE_LONG_FILE_NAMES
     strncpy(str, hostname, 14);
     str[14] = '\0';	/* in case exactly 14 chars were copied */
 #else
@@ -877,6 +951,120 @@ char *vardefn;
     return 0;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Add sysinfo(2) items as variables.  */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+void
+add_sysinfo_variables()
+{
+    char buf[500];
+    add_variable("SI_ARCHITECTURE", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_ARCHITECTURE
+    if (sysinfo(SI_ARCHITECTURE, buf, sizeof(buf)) != -1) {
+	add_variable("SI_ARCHITECTURE", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_HOSTNAME", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_HOSTNAME
+    if (sysinfo(SI_HOSTNAME, buf, sizeof(buf)) != -1) {
+	add_variable("SI_HOSTNAME", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_HW_PROVIDER", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_HW_PROVIDER
+    if (sysinfo(SI_HW_PROVIDER, buf, sizeof(buf)) != -1) {
+	add_variable("SI_HW_PROVIDER", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_HW_SERIAL", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_HW_SERIAL
+    if (sysinfo(SI_HW_SERIAL, buf, sizeof(buf)) != -1) {
+	add_variable("SI_HW_SERIAL", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_MACHINE", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_MACHINE
+    if (sysinfo(SI_MACHINE, buf, sizeof(buf)) != -1) {
+	add_variable("SI_MACHINE", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_RELEASE", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_RELEASE
+    if (sysinfo(SI_RELEASE, buf, sizeof(buf)) != -1) {
+	add_variable("SI_RELEASE", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_VERSION", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_VERSION
+    if (sysinfo(SI_VERSION, buf, sizeof(buf)) != -1) {
+	add_variable("SI_VERSION", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_SRPC_DOMAIN", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_SRPC_DOMAIN
+    if (sysinfo(SI_SRPC_DOMAIN, buf, sizeof(buf)) != -1) {
+	add_variable("SI_SRPC_DOMAIN", buf);
+    }
+#endif
+#endif
+
+    add_variable("SI_SYSNAME", "");
+#ifdef HAVE_SYSINFO
+#ifdef SI_SYSNAME
+    if (sysinfo(SI_SYSNAME, buf, sizeof(buf)) != -1) {
+	add_variable("SI_SYSNAME", buf);
+    }
+#endif
+#endif
+}
+
+#ifdef HAVE_UNAME
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Add uname(2) items as variables.  */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+void
+add_uname_variables()
+{
+    struct utsname uts;
+
+    if (uname(&uts) != -1) {
+	add_variable("UNAME_SYSNAME", uts.sysname);
+	add_variable("UNAME_NODENAME", uts.nodename);
+	add_variable("UNAME_RELEASE", uts.release);
+	add_variable("UNAME_VERSION", uts.version);
+	add_variable("UNAME_MACHINE", uts.machine);
+    } else {
+	add_variable("UNAME_SYSNAME", "");
+	add_variable("UNAME_NODENAME", "");
+	add_variable("UNAME_RELEASE", "");
+	add_variable("UNAME_VERSION", "");
+	add_variable("UNAME_MACHINE", "");
+    }
+}
+#endif
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Does variable substitution on a string.  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -884,7 +1072,8 @@ char *
 do_variables(string)
 char *string;
 {
-    char c, *dollar, *tail, *varname;
+    char c, *dollar, *varname;
+    char *tail=NULL;
     char savechar;
     char key[2];
     int l;
@@ -905,7 +1094,7 @@ char *string;
 
     variablebuf.l = 0;		/* Clear the buffer */
 
-    /* Initialize the buffer to be the same size as the input buffer */
+    /* Initialize the buffer to be at least the same size as the input buffer */
     if (!checksize(&variablebuf, ebuf.l)) {
 	Error(1, 0, "%tCouldn't increase space for variable processing\n");
 	return NULL;
@@ -999,7 +1188,7 @@ char *string;
 	    variablebuf.l += l;
 	    variablebuf.buf[variablebuf.l] = '\0';
 	    /* Restore the saved character */
-	    if (savechar)
+	    if (savechar && tail)
 		*tail = savechar;
 	}
     }
@@ -1010,6 +1199,24 @@ char *string;
     return variablebuf.buf;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* A function for printing one variable's name and value.
+ * See hprint() for calling sequence.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+void
+print_variable(indx, key, data)
+int indx;
+char *key;
+char *data;
+{
+    /* Ignore the variable "$", which is just a cute way of getting
+     * "$$" to translate to plain "$".
+     */
+    if (strcmp(key, "$") != 0)
+	printf("\t%s:\t%s\n", key, data);
+}
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Input is string like   "lhs<sep>rhs"
  * If there is no <sep>, null ptr is returned.
@@ -1071,8 +1278,8 @@ init_strqtokS()
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Finds the hostname, like gethostname(), but uses gethostbyname, if
- * available to canonicalise the name.
+/* Finds the hostname, like gethostname(), then uses canonicalize_hostname
+ * to change it to a canonical form.
  * Returns 0 on success, and -1 on error, like gethostname().
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -1081,55 +1288,70 @@ get_canonical_hostname(buf, len)
 char *buf;
 int len;
 {
-    int l;
-#ifdef USE_GETHOSTBYNAME
-    struct hostent *he, *he_nodot, *he_dot;
-    char dot_name[80] = "", nodot_name[80] = "";
-#endif
-
     if (gethostname(buf, len) < 0)
 	return Error(0, 0, "gethostname(\"%s\", %d) failed\n", buf, len);
+    
+    return canonicalize_hostname(buf, len);
+}
 
-#ifdef USE_GETHOSTBYNAME
-    if ((l=strlen(buf)) > len - 2)
-	return Error(0, 0, "hostname %s is too long for our buffer!\n", buf);
-    /* On systems using DNS, appending a dot should force gethostbyname()
-     * to fully expand and return fqdn.  On systems using NIS, appending a
-     * dot will cause match failure.  So we try both forms, and take the
-     * longer match.
-     */
-    he_nodot = gethostbyname(buf);
-    if (he_nodot && he_nodot->h_name)
-	strcpy(nodot_name, he_nodot->h_name);
-    strcat(buf, ".");
-    he_dot = gethostbyname(buf);
-    if (he_dot && he_dot->h_name)
-	strcpy(dot_name, he_dot->h_name);
-    if (strlen(dot_name) && strlen(nodot_name)) {
-	if ((strlen(dot_name) > len - 1) || (strlen(nodot_name) > len - 1))
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Canonicalizes a hostname, if available.
+ * Returns 0 on success, and -1 on error, like gethostname().
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+int
+canonicalize_hostname(buf, len)
+char *buf;
+int len;
+{
+#ifdef HAVE_GETHOSTBYNAME
+    struct hostent *he, *he_nodot, *he_dot;
+    char dot_name[80] = "", nodot_name[80] = "";
+
+
+    if (globalinfo.gethostbyname) {
+	if ((l=strlen(buf)) > len - 2 || l > sizeof(dot_name)-1)
 	    return Error(0, 0,
-			 "gethostbyname() returns `%s' which is too long for our buffer!\n",
-			 buf);
-	strcpy(buf, strlen(dot_name) > strlen(nodot_name) ?dot_name : nodot_name);
-    } else if (strlen(dot_name)) {
-	if (strlen(dot_name) > len - 1)
+			"hostname %s is too long for our buffer!\n", buf);
+	/* On systems using DNS, appending a dot should force gethostbyname()
+	 * to fully expand and return fqdn.  On systems using NIS, appending a
+	 * dot will cause match failure.  So we try both forms, and take the
+	 * longer match.
+	 */
+	he_nodot = gethostbyname(buf);
+	if (he_nodot && he_nodot->h_name)
+	    strcpy(nodot_name, he_nodot->h_name);
+	strcat(buf, ".");
+	he_dot = gethostbyname(buf);
+	if (he_dot && he_dot->h_name)
+	    strcpy(dot_name, he_dot->h_name);
+	if (strlen(dot_name) && strlen(nodot_name)) {
+	    if ((strlen(dot_name) > len - 1) || (strlen(nodot_name) > len - 1))
+	        return Error(0, 0,
+		"gethostbyname() returns `%s' which is too long for our buffer!\n",
+		    buf);
+	    strcpy(buf, strlen(dot_name) > strlen(nodot_name) ?
+							dot_name : nodot_name);
+	} else if (strlen(dot_name)) {
+	    if (strlen(dot_name) > len - 1)
+	        return Error(0, 0,
+			"gethostbyname() returns `%s' which is too long for our buffer!\n",
+			     buf);
+	    strcpy(buf, dot_name);
+	} else if (strlen(nodot_name)) {
+	    if (strlen(nodot_name) > len - 1)
+	      return Error(0, 0,
+		"gethostbyname() returns `%s' which is too long for our buffer!\n",
+			   buf);
+	    strcpy(buf, nodot_name);
+	} else {
 	    return Error(0, 0,
-			 "gethostbyname() returns `%s' which is too long for our buffer!\n",
-			 buf);
-	strcpy(buf, dot_name);
-    } else if (strlen(nodot_name)) {
-	if (strlen(nodot_name) > len - 1)
-	    return Error(0, 0,
-			 "gethostbyname() returns `%s' which is too long for our buffer!\n",
-			 buf);
-	strcpy(buf, nodot_name);
-    } else {
-	return Error(0, 0,
-		"host `%s' not recognized by gethostbyname().\n", buf);
+		    "host `%s' not recognized by gethostbyname().\n", buf);
+	}
+	/* Delete trailing dot, if still present */
+	if ((l=strlen(buf)) > 0 && buf[l-1] == '.')
+	    buf[l-1] = '\0';
     }
-    /* Delete trailing dot, if still present */
-    if ((l=strlen(buf)) > 0 && buf[l-1] == '.')
-	buf[l-1] = '\0';
 #endif
     
     return 0;
@@ -1143,26 +1365,12 @@ strtolower(s)
 char *s;
 {
     for (; *s; s++)
-	*s = tolower(*s);
+	if (isupper(*s))
+	    *s = tolower(*s);
 }
 
 
-#ifdef NEED_MEMSET
-void *
-memset(s, c, n)
-void *s;
-int c;
-int n;
-{
-    register int i;
-    register unsigned char *p = (unsigned char *) s;
-
-    for (i=0; i<n; i++)
-	*p++ = (unsigned char) c;
-}
-#endif
-
-#ifdef NEED_STRTOL
+#ifndef HAVE_STRTOL
 static unsigned long digit_a2d[128] = {0};
 
 /* Initialize the character-to-digit mapping. */
@@ -1176,7 +1384,6 @@ init_digit(void)
     for (c = '0' ; c <= '9'; c++)
 	digit_a2d[c] = c - '0';
 
-    /* C standard doesn't guarantee a-z are adjancent & increasing */
     digit_a2d['a']=digit_a2d['A'] = 10; digit_a2d['b']=digit_a2d['B'] = 11;
     digit_a2d['c']=digit_a2d['C'] = 12; digit_a2d['d']=digit_a2d['D'] = 13;
     digit_a2d['e']=digit_a2d['E'] = 14; digit_a2d['f']=digit_a2d['F'] = 15;
@@ -1248,19 +1455,36 @@ strtol(register const char *str, char **ptr, int base)
 }
 #endif
 
-#ifdef NEED_STRERROR
+#ifndef HAVE_MEMSET
+void *
+memset(s, c, n)
+void *s;
+int c;
+int n;
+{
+    register int i;
+    register unsigned char *p = (unsigned char *) s;
 
-#ifndef NEED_SYS_ERRLIST
+    for (i=0; i<n; i++)
+	*p++ = (unsigned char) c;
+}
+#endif
+
+#ifdef HAVE_SYS_ERRLIST
     extern char *sys_errlist[];
     extern int sys_nerr;
 #endif
 
+/*
+ * Safe strerror(): range-checks the argument.
+ * (C standard doesn't guarantee any range checks.)
+ */
 char
-*strerror(errnum)
+*Strerror(errnum)
 int errnum;
 {
     static char buf[500];
-#ifdef NEED_SYS_ERRLIST
+#ifndef HAVE_SYS_ERRLIST
     sprintf(buf, "Error %d", errnum);
     return buf;
 #else
@@ -1268,9 +1492,55 @@ int errnum;
 	sprintf(buf, "Error %d (!)", errnum);
 	return buf;
     } else {
+#ifdef HAVE_STRERROR
+	return strerror(errnum);
+#else
 	return sys_errlist[errnum];
+#endif
     }
 #endif
 }
 
+#ifndef HAVE_STRDUP
+char *
+strdup(s)
+char *s;
+{
+    char *t = (char *) malloc(strlen(s)+1);
+    strcpy(t, s);
+    return t;
+}
+#endif
+
+#ifndef HAVE_VPRINTF
+/* Very system-specific.  May not work for you...
+ */
+void
+vfprintf(stream, format, ap)
+FILE *stream;
+char *format;
+va_list ap;
+{
+    _doprnt(format, ap, stream);
+}
+
+void
+vsprintf(str, format, ap)
+char *str;
+char *format;
+va_list ap;
+{
+    FILE fp;
+    fp._cnt = 077777;
+    fp._base = fp._ptr = str;
+    fp._bufsiz = 0;
+    fp._flag = _IOSTRG | _IOWRT;
+    fp._file = 0;
+
+    _doprnt(format, ap, &fp);
+}
+#ifndef HAVE_DOPRNT
+ERROR -- YOU HAVE NEITHER v?printf() NOR _doprnt().
+YOU ARE HOSED UNTIL YOU SUPPLY ONE OF THESE.
+#endif
 #endif
